# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_route_table_manager
   :synopsis: module that defines several test cases for the
              route_table_manager module.
   In particular, unit tests for RouteTableManager class.
   Setup : Start RouteTableManager thread instance.
   TearDown : Stop RouteTableManager thread instance.
   RouteTableManager is in charge to maintain the list of Workers
   (TrackerWorker, ExaBGPPeerWorker, BGPPeerWorker) subscriptions, to
   process BGB route events and to dispatch BGP routes to the Workers according
   to their subscriptions.
   Mock class is used to stub Workers.
   Tests are organized as follow :
   - testAx use cases to test worker subscriptions to route target (or match)
     1- with no route to synthesize
     2- with routes to synthesize <=> advertise existing routes to the new
        worker subscription according following rules :
        - route should not be synthesized to its source,
        - route should not be synthesized between BGPPeerWorkers
        - route should not be synthesized if already send to worker for
          another RT
     For both use cases check that (worker, match) is correctly recorded by
     RouteTableManager
     Other test cases : re-subscription to check routes are not synthesized
   - testBx use cases to test worker unsubscriptions to match (without and with
     routes to synthesize) :
     same rules should be applied to generate withdraw events.
     Check (worker, match) has been deleted
     Other test cases : unsubscription to match or by worker not registered
   - testCx to test the processing of route event generated by a worker.
     2 types of route event : advertise (new or update) or withdraw
     1- advertise or withdraw a route entry without event propagation to
        workers: check that entry is recorded/deleted
     2- advertise (new or update) or withdraw with event propagation:
        to test dispatching of route events to the workers according to
        their subscriptions.
     Other test cases : withdraw of a not registered route, advertise of the
     same route (same attr and RTs)
   - testDx : to test worker cleanup
   - testEx : to test dumpState

"""

import logging

import mock

from testtools import TestCase

from bagpipe.bgp.tests import BaseTestBagPipeBGP, RT1, RT2, RT3, \
    NLRI1, NLRI2, NH1, NH2

from bagpipe.bgp.engine import RouteEvent
from bagpipe.bgp.engine import RouteEntry
from bagpipe.bgp.engine import Subscription
from bagpipe.bgp.engine import Unsubscription
from bagpipe.bgp.engine.worker import Worker
from bagpipe.bgp.engine.bgp_peer_worker import BGPPeerWorker
from bagpipe.bgp.engine.route_table_manager import RouteTableManager
from bagpipe.bgp.engine.route_table_manager import Match
from bagpipe.bgp.engine.route_table_manager import WorkerCleanupEvent

from bagpipe.exabgp.message.update.attributes import Attributes
from bagpipe.exabgp.structure.address import AFI, SAFI

log = logging.getLogger()

MATCH1 = Match(AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn), RT1)
MATCH2 = Match(AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn), RT2)
MATCH3 = Match(AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn), RT3)


class TestRouteTableManager(TestCase, BaseTestBagPipeBGP):

    def setUp(self):
        super(TestRouteTableManager, self).setUp()
        self.routeTableManager = RouteTableManager()
        self.routeTableManager.start()
        self.setEventTargetWorker(self.routeTableManager)

    def tearDown(self):
        super(TestRouteTableManager, self).tearDown()
        self.routeTableManager.stop()
        self.routeTableManager.join()

    def _newworker(self, workerName, workerType):
        worker = mock.Mock(spec=workerType)
        worker.name = workerName
        worker.enqueue = mock.Mock()
        return worker

    def _workerSubscriptions(self, worker, rts,
                             afi=AFI(AFI.ipv4), safi=SAFI(SAFI.mpls_vpn)):
        for rt in rts:
            subscribe = Subscription(afi, safi, rt, worker)
            self.routeTableManager.enqueue(subscribe)

    def _workerUnsubscriptions(self, worker, rts,
                               afi=AFI(AFI.ipv4), safi=SAFI(SAFI.mpls_vpn)):
        for rt in rts:
            unsubscribe = Unsubscription(afi, safi, rt, worker)
            self.routeTableManager.enqueue(unsubscribe)

    def _checkSubscriptions(self, worker, matches):
        for match in matches:
            subs = self.routeTableManager.getWorkerSubscriptions(worker)
            self.assertIn(match, subs, "Subscription not found")

    def _checkUnsubscriptions(self, worker, matches):
        for match in matches:
            subs = self.routeTableManager.getWorkerSubscriptions(worker)
            self.assertNotIn(match, subs, "Subscription not found")

    def _checkCalls(self, worker, entry):
        self.assertTrue(
            entry in self.routeTableManager.getWorkerRouteEntries(worker),
            "Route entry not found")

    def _checkNoRouteEntry(self, worker, entry):
        self.assertTrue(
            entry not in self.routeTableManager.getWorkerRouteEntries(worker),
            "Route entry found")

    def _checkEventsCalls(self, events, advertisedRoutes, withdrawnNLRIs):
        '''
        checks that each advertise event in 'events' is in advertisedRoutes,
        that each withdraw event in 'events' is in withdrawnNLRIs
        and that all events in withdrawnNLRIs and advertisedRoutes are in
        'events'
        '''
        for (callArgs, _) in events:
            if (callArgs[0].type == RouteEvent.ADVERTISE):
                self.assertIn(callArgs[0].routeEntry, advertisedRoutes,
                              "Bad advertised route")
                advertisedRoutes.remove(callArgs[0].routeEntry)
            else:  # WITHDRAW
                self.assertIn(callArgs[0].routeEntry.nlri, withdrawnNLRIs,
                              "Bad withdrawn route")
                withdrawnNLRIs.remove(callArgs[0].routeEntry.nlri)
        self.assertEqual(0, len(advertisedRoutes), "some routes not advert'd")
        self.assertEqual(0, len(withdrawnNLRIs), "some routes not withdrawn")

    def testA1_SubscriptionsWithNoRouteTosynthesize(self):
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # Waiting for RouteTableManager thread finishes to process the
        # subscriptions
        self._wait()
        # check subscriptions
        self._checkSubscriptions(worker1, [MATCH1, MATCH2])

    def testA2_SubscriptionsWithRouteTosynthesize(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                   [RT1, RT2], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI2,
                                   [RT2], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._workerSubscriptions(bgpPeerWorker1, [RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # Worker2 subscribes to RT1
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT1])
        # Worker3 subscribes to RT3
        worker3 = self._newworker("Worker-3", Worker)
        self._workerSubscriptions(worker3, [RT3])
        # BGPPeerWorker2 subscribes to RT1
        bgpPeerWorker2 = self._newworker("BGPWorker2", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker2, [RT1])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check route entry synthesized
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be synthesized to its source")
        self.assertEqual(0, worker3.enqueue.call_count,
                         "no route should be synthesized to Worker3")
        self.assertEqual(0, bgpPeerWorker2.enqueue.call_count,
                         "Route should not be synthesized between BGP workers")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 advertise events should be synthesized to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry], [])
        self.assertEqual(1, worker2.enqueue.call_count,
                         "1 advertise event should be synthesized to Worker2")
        self._checkEventsCalls(
            worker2.enqueue.call_args_list, [evt1.routeEntry], [])

    def testA3_ReSubscription(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgpPeerWorker1 = self._newworker(
            "BGPWorker1", BGPPeerWorker)
        routeEvent = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                         [RT1, RT2], bgpPeerWorker1, NH1)
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # Worker1 subscribes again to RT1
        self._workerSubscriptions(worker1, [RT1])
        # Worker2 subscribes to RT1
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT1])
        # Worker1 subscribes again to RT2
        self._workerSubscriptions(worker2, [RT2])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check route entry synthesized
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [routeEvent.routeEntry], [])
        self.assertEqual(1, worker2.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker2")
        self._checkEventsCalls(worker2.enqueue.call_args_list,
                               [routeEvent.routeEntry], [])

    def testB1_UnsubscriptionWithNoRouteTosynthesize(self):
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker1, [RT1, RT2])
        # Worker1 unsubscribes to RT1
        self._workerUnsubscriptions(worker1, [RT1])
        # BGPPeerWorker1 unsubscribes to RT1 and RT2
        self._workerUnsubscriptions(bgpPeerWorker1, [RT1, RT2])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check subscription/unsubscriptions
        self._checkUnsubscriptions(worker1, [MATCH1])
        self._checkSubscriptions(worker1, [MATCH2])
        self._checkUnsubscriptions(bgpPeerWorker1, [MATCH1, MATCH2])

    def testB2_UnsubscriptionWithRouteTosynthesize(self):
        # BGPPeerWorker1 advertises a route for RT1
        bgpPeerWorker1 = self._newworker(
            "BGPWorker1", BGPPeerWorker)
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                                   bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI2, [RT2],
                                   bgpPeerWorker1, NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._workerSubscriptions(bgpPeerWorker1, [RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # Worker2 subscribes to RT2
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT2])
        # Worker3 subscribes to RT3
        worker3 = self._newworker("Worker-3", Worker)
        self._workerSubscriptions(worker3, [RT3])
        # BGPPeerWorker2 subscribes to RT1
        bgpPeerWorker2 = self._newworker("BGPWorker2", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker2, [RT1])
        # Workers and BGPPeerWorker unsubscriptions
        self._workerUnsubscriptions(bgpPeerWorker1, [RT1])
        self._workerUnsubscriptions(worker1, [RT1])
        self._workerUnsubscriptions(worker2, [RT2])
        self._workerUnsubscriptions(worker3, [RT3])
        self._workerUnsubscriptions(bgpPeerWorker2, [RT1])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check route entry synthesized
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be synthesized to its source")
        self.assertEqual(0, worker3.enqueue.call_count,
                         "no route should be synthesized to Worker3")
        self.assertEqual(0, bgpPeerWorker2.enqueue.call_count,
                         "Route should not be synthesized between "
                         "BGPPeerWorkers")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 advertise event should be synthesized to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry], [])
        self.assertEqual(4, worker2.enqueue.call_count,
                         "4 events should be synthesized to Worker2: "
                         "2 advertise and 2 withdraw")
        self._checkEventsCalls(worker2.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry],
                               [evt1.routeEntry.nlri, evt2.routeEntry.nlri])

    def testB3_UnsubscriptionNotRegistered(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1])
        # Worker1 unsubscribes to RT2
        self._workerUnsubscriptions(worker1, [RT2])
        # BGPPeerWorker1 unsubscribes to RT1
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        self._workerUnsubscriptions(bgpPeerWorker1, [RT1, RT2])
        # Waiting for RouteTableManager thread finishes to process the
        # subscription
        self._wait()
        # check subscription/unsubscriptions
        self._checkSubscriptions(worker1, [MATCH1])
        self._checkUnsubscriptions(bgpPeerWorker1, [MATCH1, MATCH2])

    def testC1_routeAdvertiseByWorkerWithoutPropagation(self):
        # Worker1 advertises a route for RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        routeEvent = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                         [RT1, RT2], worker1, NH1)
        # check route entry has been inserted
        self._checkCalls(worker1, routeEvent.routeEntry)

    def testC2_routeWithdrawByWorkerWithoutPropagation(self):
        # Worker1 advertises then withdraws a route
        worker1 = self._newworker("Worker-1", Worker)
        self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                            [RT1, RT2], worker1, NH1)
        routeEvent = self._newRouteEvent(RouteEvent.WITHDRAW, NLRI1,
                                         [RT1], worker1, NH1)
        # check route entry has been removed
        self._checkNoRouteEntry(worker1, routeEvent.routeEntry)

    def testC3_routeAdvertiseByBGPPeerWithPropagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1])
        # Worker2 subscribes to RT2
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker1, [RT1, RT2])
        # BGPPeerWorker2 subscribes to RT1 and RT2
        bgpPeerWorker2 = self._newworker("BGPWorker2", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker2, [RT1, RT2])
        # BGPPeerWorker1 advertises a route for RT1
        routeEvent = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                         [RT1], bgpPeerWorker1, NH1)
        # check routeEvent propagation
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route should be propagated to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [routeEvent.routeEntry], [])
        self.assertEqual(0, worker2.enqueue.call_count,
                         "no route should be propagated to Worker2")
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(0, bgpPeerWorker2.enqueue.call_count,
                         "Route should not be propagated between BGP workers")

    def testC4_routeWithdrawByPeerWorkerWithPropagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1])
        # Worker2 subscribes to RT2
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT2])
        # Worker3 subscribes to RT3
        worker3 = self._newworker("Worker-3", Worker)
        self._workerSubscriptions(worker3, [RT3])
        # BGPPeerWorker1 subscribes to RT1
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker1, [RT1])
        # BGPPeerWorker2 subscribes to RT2
        bgpPeerWorker2 = self._newworker("BGPWorker2", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker2, [RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        routeEventA = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                          [RT1, RT2], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 withdraw previous route (without RT
        routeEventW = self._newRouteEvent(RouteEvent.WITHDRAW, NLRI1,
                                          [], bgpPeerWorker1, NH1)
        # check routeEvent propagation
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be propagated to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [routeEventA.routeEntry],
                               [routeEventW.routeEntry.nlri])
        self.assertEqual(2, worker2.enqueue.call_count,
                         "2 routes should be propagated to Worker2")
        self._checkEventsCalls(worker2.enqueue.call_args_list,
                               [routeEventA.routeEntry],
                               [routeEventW.routeEntry.nlri])
        self.assertEqual(0, worker3.enqueue.call_count,
                         "No route should be propagated to Worker3")
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(0, bgpPeerWorker2.enqueue.call_count,
                         "Route should not be propagated between BGP workers")

    def testC5_routeUpdateByBGPPeerWithWithdrawPropagation(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1])
        # Worker2 subscribes to RT2
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT2])
        # Worker3 subscribes to RT3
        worker3 = self._newworker("Worker-3", Worker)
        self._workerSubscriptions(worker3, [RT3])
        # BGPPeerWorker1 advertises a route for RT1, RT2 and RT3
        bgpPeerWorker1 = self._newworker(
            "BGPWorker1", BGPPeerWorker)
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                   [RT1, RT2, RT3], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises the same nlri with attributes NH and RTs
        # modification
        evt2 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                                   [RT1, RT2], bgpPeerWorker1, NH2)
        # check route event propagation
        # TO DO : check routeEvent.replacedRoute
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be advertised to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry], [])
        self.assertEqual(2, worker2.enqueue.call_count,
                         "2 routes should be advertised to Worker2")
        self._checkEventsCalls(worker2.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry], [])
        self.assertEqual(2, worker3.enqueue.call_count,
                         "2 routes should be advert/withdraw to Worker3")
        self._checkEventsCalls(worker3.enqueue.call_args_list,
                               [evt1.routeEntry], [evt1.routeEntry.nlri])

    def testC6_routeReadvertised(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2, RT3])
        # BGPPeerWorker1 advertises a route for RT1, RT2 and RT3
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                                   bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises the same nlri with same attributes and RTs
        self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                            bgpPeerWorker1, NH1)
        # check route event propagation
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be propagated to its source")
        self.assertEqual(1, worker1.enqueue.call_count,
                         "only 1 route should be advertised to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry], [])

    def testC7_routeWithdrawNotRegistered(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgpPeerWorker1 = self._newworker(
            "BGPWorker1", BGPPeerWorker)
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                                   bgpPeerWorker1, NH1)
        # BGPPeerWorker1 withdraw a not registered route (without RT
        self._newRouteEvent(RouteEvent.WITHDRAW, NLRI2, [],
                            bgpPeerWorker1, NH1)
        # Waiting for RouteTableManager thread finishes to process routeEvent
        self._wait()
        # check routeEvent propagation
        self.assertEqual(1, worker1.enqueue.call_count,
                         "1 route1 should be propagated to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry], [])
        self.assertEqual(0, bgpPeerWorker1.enqueue.call_count,
                         "Route should not be propagated back to its source")

    def testD1_WorkerCleanup(self):
        # Worker1 subscribes to RT1
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1])
        # Worker2 subscribes to RT2
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT2])
        # BGPPeerWorker1 subscribes to RT1 and RT2
        bgpPeerWorker1 = self._newworker(
            "BGPWorker1", BGPPeerWorker)
        self._workerSubscriptions(bgpPeerWorker1, [RT1, RT2])
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        evt1 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1, [RT1, RT2],
                                   bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        evt2 = self._newRouteEvent(RouteEvent.ADVERTISE, NLRI2, [RT2],
                                   bgpPeerWorker1, NH1)
        # Cleanup Worker1
        self.routeTableManager.enqueue(WorkerCleanupEvent(bgpPeerWorker1))
        # Waiting for RouteTableManager thread finishes to process the
        # subscriptions
        self._wait()
        # check unsubscriptions
        self._checkUnsubscriptions(bgpPeerWorker1, [MATCH1, MATCH2])
        # Check route synthesize to Worker1 and Worker2
        self.assertEqual(2, worker1.enqueue.call_count,
                         "2 routes should be advert/withdraw to Worker1")
        self._checkEventsCalls(worker1.enqueue.call_args_list,
                               [evt1.routeEntry], [evt1.routeEntry.nlri])
        self.assertEqual(4, worker2.enqueue.call_count,
                         "4 routes should be advert/withdraw to Worker2")
        self._checkEventsCalls(worker2.enqueue.call_args_list,
                               [evt1.routeEntry, evt2.routeEntry],
                               [evt1.routeEntry.nlri, evt2.routeEntry.nlri])
        # Check route entries have been removed for BGPPeerWorker1
        self._checkNoRouteEntry(bgpPeerWorker1, evt1.routeEntry)
        self._checkNoRouteEntry(bgpPeerWorker1, evt2.routeEntry)

    def testE1_DumpState(self):
        # BGPPeerWorker1 advertises a route for RT1 and RT2
        bgpPeerWorker1 = self._newworker("BGPWorker1", BGPPeerWorker)
        self._newRouteEvent(RouteEvent.ADVERTISE, NLRI1,
                            [RT1, RT2], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 advertises an other route for RT2
        self._newRouteEvent(RouteEvent.ADVERTISE, NLRI2,
                            [RT2], bgpPeerWorker1, NH1)
        # BGPPeerWorker1 subscribes to RT1
        self._workerSubscriptions(bgpPeerWorker1, [RT1])
        # Worker1 subscribes to RT1 and RT2
        worker1 = self._newworker("Worker-1", Worker)
        self._workerSubscriptions(worker1, [RT1, RT2])
        # Worker2 subscribes to RT1
        worker2 = self._newworker("Worker-2", Worker)
        self._workerSubscriptions(worker2, [RT1])
        # Worker3 subscribes to RT3
        worker3 = self._newworker("Worker-3", Worker)
        self._workerSubscriptions(worker3, [RT3])

        # Waiting for RouteTableManager thread finishes to process the
        # subscriptions and route event processing
        self._wait()

        self.routeTableManager._dumpState()

    def testF1_testEmptyRT(self):
        # worker advertises a route with no RT

        w1 = self._newworker("Worker1", Worker)

        subscribe = Subscription(AFI(AFI.ipv4),
                                 SAFI(SAFI.mpls_vpn),
                                 None, w1)
        self.routeTableManager.enqueue(subscribe)

        w2 = self._newworker("Worker2", Worker)

        routeEvent = RouteEvent(RouteEvent.ADVERTISE, RouteEntry(
            AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn), None, NLRI1, Attributes(), w2), w2)

        self.routeTableManager.enqueue(routeEvent)

        self._wait()

        self.assertEqual(1, w1.enqueue.call_count,
                         "1 route advertised should be synthesized to Worker1")



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback

from threading import Thread
from Queue import Queue

from bagpipe.bgp.engine import RouteEvent, Subscription, Unsubscription
from bagpipe.bgp.engine.worker import Worker
from bagpipe.bgp.engine.bgp_peer_worker import BGPPeerWorker

from bagpipe.bgp.common.looking_glass import LookingGlass, LGMap
from bagpipe.bgp.common import logDecorator

from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.message.update.attribute.communities import RouteTarget

log = logging.getLogger(__name__)


class WorkerCleanupEvent(object):

    def __init__(self, worker):
        self.worker = worker

    def __repr__(self):
        return "WorkerCleanupEvent:%s" % (self.worker.name)


class Match(object):

    def __init__(self, afi, safi, routeTarget):
        assert(isinstance(afi, AFI))
        assert(isinstance(safi, SAFI))
        assert(routeTarget is None or isinstance(routeTarget, RouteTarget))
        self.afi = afi
        self.safi = safi
        self.routeTarget = routeTarget

    # FIXME: use a better hash if needed for performances
    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return "match:%s/%s,%s" % (self.afi or "*", self.safi or "*",
                                   self.routeTarget or "*")

    def __cmp__(self, other):
        assert isinstance(other, Match)

        self_afi = self.afi or AFI(0)
        self_safi = self.safi or SAFI(0)
        self_rt = self.routeTarget or RouteTarget(0, None, 0)

        other_afi = other.afi or AFI(0)
        other_safi = other.safi or SAFI(0)
        other_rt = other.routeTarget or RouteTarget(0, None, 0)

        return cmp((self_afi,  self_safi,  self_rt),
                   (other_afi, other_safi, other_rt))

StopEvent = "StopEvent"


class RouteTableManager(Thread, LookingGlass):

    """
    This singleton class will dispatch events between Workers.
    Events relates to BGP routes that are announced or withdrawn by workers.
    Workers subscribe to events by indicating AFI, SAFI, Route Targets in which
    they are interested.
    There can be workers responsible for handling services locally (e.g. a VPN
    instance) and workers that are BGP peers.

    Though sources of routes are typically Workers, they don't need to be; any
    class can source a route.
    """
    class WorkersAndEntries(object):

        def __init__(self):
            self.workers = set()
            self.entries = set()

        def __repr__(self):
            return "workers: %s\nentries: %s" % (self.workers,
                                                 self.entries)

    def __init__(self):
        Thread.__init__(self, name="RouteTableManager")
        self.setDaemon(True)

        self._match2workersAndEntries = {}
        # keys are Matches, values are WorkersAndEntries objects
        self._worker2matches = {}  # keys are Workers, values are Match objects
        self._source_nlri2entry = {}
        # keys are (source,nlri) tuples, values are Entry objects
        self._source2entries = {}
        # dict: keys are event sources, each value is a set() of Entry
        # objects

        self._queue = Queue()

    @logDecorator.logInfo
    def stop(self):
        self.enqueue(StopEvent)

    def run(self):
        while True:
            log.debug("RouteTableManager waiting on queue")
            event = self._queue.get()
            log.debug("RouteTableManager received event %s", event)
            try:
                if event.__class__ == RouteEvent:
                    self._receiveRouteEvent(event)
                elif event.__class__ == Subscription:
                    self._workerSubscribes(event)
                elif event.__class__ == Unsubscription:
                    self._workerUnsubscribes(event)
                elif event.__class__ == WorkerCleanupEvent:
                    self._workerCleanup(event.worker)
                elif event == StopEvent:
                    log.info("StopEvent => breaking main loop")
                    break
            except Exception as e:
                log.error("Exception during processing of event: %s", repr(e))
                log.error("    event was: %s", event)
                log.error("%s", traceback.format_exc())

            log.debug("RouteTableManager queue size: %d", self._queue.qsize())

        log.info("Out of main loop")

    def enqueue(self, event):
        self._queue.put(event)

    def _checkMatch2workersAndEntriesCleanup(self, match):
        try:
            item = self._match2workersAndEntries[match]
        except KeyError:
            log.warning("why are we here ?")
            # nothing to cleanup
            return

        if len(item.workers) == 0 and len(item.entries) == 0:
            del self._match2workersAndEntries[match]

    def _match2workersAndEntriesLookupCreate(self, match):
        try:
            return self._match2workersAndEntries[match]
        except KeyError:
            wa = RouteTableManager.WorkersAndEntries()
            self._match2workersAndEntries[match] = wa
            return wa

    def _match2entries(self, match, createIfNone=False, emptyListIfNone=True):
        if createIfNone:
            return self._match2workersAndEntriesLookupCreate(match).entries
        try:
            return self._match2workersAndEntries[match].entries
        except KeyError:
            if emptyListIfNone:
                return []
            else:
                raise

    def _match2workers(self, match, createIfNone=False, emptyListIfNone=True):
        if createIfNone:
            return self._match2workersAndEntriesLookupCreate(match).workers
        try:
            return self._match2workersAndEntries[match].workers
        except KeyError:
            if emptyListIfNone:
                return []
            else:
                raise

    def _source2entriesAddEntry(self, entry):
        try:
            entries = self._source2entries[entry.source]
        except KeyError:
            entries = set()
            self._source2entries[entry.source] = entries
        entries.add(entry)

    def _source2entriesRemoveEntry(self, entry):
        try:
            entries = self._source2entries[entry.source]
            log.debug("_source2entries[ entry.worker ] = %s ", entries)
            entries.discard(entry)
        except KeyError:
            log.debug("(attempt at removing a non existing entry from "
                      "self._source2entries[ entry.source ] : %s)", entry)
            pass

    def getWorkerSubscriptions(self, worker):
        """
        the returned entries should *not* be modified by caller
        (TODO: protect worker object by making their variables private)
        """

        if worker not in self._worker2matches:
            return []
        else:
            matches = self._worker2matches[worker]
            if matches is None:
                return []
            else:
                return sorted(matches)

    def getWorkerRouteEntries(self, worker):
        """
        the returned entries should *not* be modified by caller
        (TODO: protect Entry objects by making their variables private)
        """
        if worker not in self._source2entries:
            return []
        else:
            entries = self._source2entries[worker]
            if entries is None:
                return []
            else:
                return entries

    def _workerSubscribes(self, sub):
        # TODO: this function currently will not consider whether or not
        # is already subscribed to set of route events, before considering
        # a subscription.  In particular, multiple identical subscriptions
        # will lead to this code resyntethizing events at each call.
        #
        # Ideally, the code should detect that the worker is already subscribed
        # and skip the subscription. *But* such a change should not be done
        # until the code in ExaBGPPeerWorker is updated to support this.

        assert(isinstance(sub.worker, Worker))
        log.info("workerSubscribes: %s", sub)

        worker = sub.worker

        # self._dumpState()

        match = Match(sub.afi, sub.safi, sub.routeTarget)

        # update match2worker
        self._match2workers(match, createIfNone=True).add(worker)

        # update worker2matches
        if worker not in self._worker2matches:
            self._worker2matches[worker] = set()

        # re-synthesize events
        for entry in self._match2entries(match):
            log.debug("Found a entry for this match: %s", entry)
            event = RouteEvent(RouteEvent.ADVERTISE, entry)
            (shouldDispatch, reason) = self._shouldDispatch(event, worker)

            if shouldDispatch:
                # check if the entry carries a routeTarget to which the worker
                # was already subscribed
                for rt in entry.routeTargets:
                    if Match(entry.afi,
                             entry.safi,
                             rt) in self._worker2matches[worker]:
                        (shouldDispatch, reason) = (
                            False,
                            "worker already had a subscription for this route")
                        break

            if shouldDispatch:
                log.info("Dispatching re-synthesized event for %s", entry)
                worker.enqueue(event)
            else:
                log.info("%s => not dispatching re-synthesized event for %s",
                         reason, entry)

        # update worker2matches
        self._worker2matches[worker].add(match)

        # self._dumpState()

    def _workerUnsubscribes(self, sub):
        assert(isinstance(sub.worker, Worker))

        # self._dumpState()

        match = Match(sub.afi, sub.safi, sub.routeTarget)

        # update _worker2matches
        if sub.worker not in self._worker2matches:
            log.warning("worker %s unsubs'd from %s but wasn't tracked yet",
                        sub.worker, match)
        else:
            try:
                self._worker2matches[sub.worker].remove(match)
            except KeyError:
                log.warning("worker %s unsubs' from %s but this match was"
                            "not tracked for this worker (should not happen,"
                            " this is a bug)", sub.worker, match)

        # synthesize withdraw events
        for entry in self._match2entries(match, emptyListIfNone=True):
            intersect = set(
                self._matchesFor(entry.afi, entry.safi, entry.routeTargets)
                ).intersection(self._worker2matches[sub.worker])
            if len(intersect) > 0:
                log.debug("Will not synthesize withdraw event for %s, because"
                          " worker subscribed to %s", entry, intersect)
            else:
                log.debug("Found a entry for this match: %s", entry)
                event = RouteEvent(RouteEvent.WITHDRAW, entry)
                (shouldDispatch, reason) = self._shouldDispatch(event,
                                                                sub.worker)
                if shouldDispatch:
                    log.info("Dispatching re-synthesized event for %s", entry)
                    sub.worker.enqueue(event)
                else:
                    log.info(
                        "%s => not dispatching re-synthesized event for %s",
                        reason, entry)

        # update _match2workersAndEntries
        if match not in self._match2workersAndEntries:
            log.warning("worker %s unsubscribed from %s but we had no such"
                        " subscription yet", sub.worker, match)
        else:
            try:
                self._match2workers(match).remove(sub.worker)
            except KeyError:
                log.warning("worker %s unsubscribed from %s but was not"
                            " subscribed yet", sub.worker, match)

        self._checkMatch2workersAndEntriesCleanup(match)

        # self._dumpState()

    def _matchesFor(self, afi, safi, routeTargets):
        # generate all possible match entries for this afi/safi
        # and these routetargets, with all possible wildcards
        #
        # There are 4*(n+1) possible Match object (for n routeTargets)
        for _afi in (Subscription.ANY_AFI, afi):
            for _safi in (Subscription.ANY_SAFI, safi):
                yield Match(_afi, _safi, None)
                if routeTargets is not None:
                    for rt in routeTargets:
                        yield Match(_afi, _safi, rt)

    def _propagateRouteEvent(self, routeEvent, exceptWorkers=None):
        '''Propagate routeEvent to workers subscribed to the route RTs
        or wildcards, except the workers in exceptWorkers. Returns the list of
        workers to which the event was propagated.'''

        log.debug("Propagate event to interested workers: %s", routeEvent)

        re = routeEvent.routeEntry

        if exceptWorkers is None:
            exceptWorkers = []

        targetWorkers = set()
        for match in self._matchesFor(re.afi, re.safi, re.routeTargets):
            log.debug("Finding interested workers for match %s", match)

            interestedWorkers = self._match2workers(match,
                                                    emptyListIfNone=True)
            log.debug("   Workers interested in this match: %s",
                      interestedWorkers)

            for worker in interestedWorkers:
                (shouldDispatch, reason) = self._shouldDispatch(routeEvent,
                                                                worker)
                if shouldDispatch:
                    if worker not in exceptWorkers:
                        log.debug("Will dispatch event to %s: %s",
                                  worker, routeEvent)
                        targetWorkers.add(worker)
                    else:
                        log.debug("Decided not to dispatch to %s, based on "
                                  "exceptWorkers: %s", worker, routeEvent)
                else:
                    log.debug("Decided not to dispatch to %s: %s (%s)",
                              worker, reason, routeEvent)

        for worker in targetWorkers:
            log.info("Dispatching event to %s: %s", worker, routeEvent)
            worker.enqueue(routeEvent)

        return targetWorkers

    def _receiveRouteEvent(self, routeEvent):
        log.info("receive: %s", routeEvent)

        entry = routeEvent.routeEntry

        log.debug("Try to find a entry from same peer with same nlri")
        try:
            replacedEntry = self._source_nlri2entry[(entry.source, entry.nlri)]
        except KeyError:
            replacedEntry = None

        log.debug("   Result: %s", replacedEntry)

        # replacedEntry should be non-empty for a withdraw
        if replacedEntry is None and (routeEvent.type == RouteEvent.WITHDRAW):
            log.warning("WITHDRAW but found no route that we could remove: %s",
                        routeEvent.routeEntry)
            return

        # Propagate events to interested workers...
        if routeEvent.type == RouteEvent.ADVERTISE:

            if replacedEntry == routeEvent.routeEntry:
                log.warning("The route advertized is the same as the one "
                            "previously advertized by the source, ignoring")
                return

            # propagate event to interested worker
            # and include the info on the route are replaced by this
            # route, if any
            routeEvent.setReplacedRoute(replacedEntry)

            workersAlreadyNotified = self._propagateRouteEvent(routeEvent)
        else:  # WITHDRAW
            workersAlreadyNotified = None

        # Synthesize and dispatch a withdraw event for the route entry that
        # was withdrawn or replaced, except, in the case of a replaced route,
        # to workers that had the ADVERTISE event
        if replacedEntry is not None:
            log.debug("Synthesizing a withdraw event for replaced route %s",
                      replacedEntry)
            removalEvent = RouteEvent(RouteEvent.WITHDRAW,
                                      replacedEntry,
                                      routeEvent.source)

            self._propagateRouteEvent(removalEvent, workersAlreadyNotified)

            # Update match2entries and source2entries for the
            # replacedRoute
            for match in self._matchesFor(replacedEntry.afi,
                                          replacedEntry.safi,
                                          replacedEntry.routeTargets):
                try:
                    self._match2entries(match).discard(replacedEntry)
                except KeyError:
                    log.error("Trying to remove a route from a match, but"
                              " match %s not found - not supposed to happen"
                              " (route: %s)", match, replacedEntry)
                self._checkMatch2workersAndEntriesCleanup(match)

            self._source2entriesRemoveEntry(replacedEntry)

        if routeEvent.type == RouteEvent.ADVERTISE:
            # Update match2entries and source2entries for the newly
            # advertized route
            for match in self._matchesFor(entry.afi,
                                          entry.safi,
                                          entry.routeTargets):
                self._match2entries(match, createIfNone=True).add(entry)

            self._source2entriesAddEntry(entry)

            # Update _source_nlri2entry
            self._source_nlri2entry[(entry.source, entry.nlri)] = entry
        else:  # WITHDRAW
            # Update _source_nlri2entry
            try:
                del self._source_nlri2entry[(entry.source, entry.nlri)]
            except KeyError:
                log.error("Withdraw, but nothing removed in "
                          "_sourcenlri2entryRemove")

        #  self._dumpState()

    def _shouldDispatch(self, routeEvent, targetWorker):
        '''
        returns a (boolean,string) tuple
        the string contains the reason why the routeEvent should not be
        dispatched to targetWorker
        '''
        if (routeEvent.source == targetWorker):
            return (False, "not dispatching an update back to its source")
        elif (isinstance(routeEvent.source, BGPPeerWorker)
              and isinstance(targetWorker, BGPPeerWorker)):
            return (False, "do not dispatch a route between BGP peers")
        else:
            return (True, "")

    def _workerCleanup(self, worker):
        '''
        Consider all routes announced by this worker as withdrawn.
        Consider this worker unsubscribed from all of its current
        subscriptions.
        '''
        log.info("Cleanup for worker %s", worker.name)
        # synthesize withdraw events for all routes from this worker
        if worker in self._source2entries:
            entries = self._source2entries[worker]
            log.info("  Preparing to withdraw %d routes that were advertised "
                     "by worker", len(entries))
            for entry in entries:
                log.info("  Enqueue event to Withdraw route %s", entry)
                self.enqueue(RouteEvent(RouteEvent.WITHDRAW, entry))

            del self._source2entries[worker]
        else:
            log.info("(we had no trace of %s in _source2entries)", worker)

        # remove worker from all of its subscriptions
        if worker in self._worker2matches:
            for match in self._worker2matches[worker]:
                assert(match in self._match2workersAndEntries)
                self._match2workers(match).remove(worker)
            del self._worker2matches[worker]

        # self._dumpState()

    def _dumpState(self):
        if not log.isEnabledFor(logging.DEBUG):
            return

        dump = []

        dump.append("~~~ Worker -> Matches ~~~")
        for worker in self._worker2matches.keys():
            dump.append("  %s" % worker)
            matches = list(self._worker2matches[worker])
            matches.sort()
            for match in matches:
                dump.append("    %s" % match)

        dump.append("\n~~~ Source -> Entries ~~~")
        for source in self._source2entries.keys():
            dump.append("  %s" % source)
            for entry in self._source2entries[source]:
                dump.append("    %s" % entry)

        match2workerDump = []
        match2entriesDump = []

        matches = list(self._match2workersAndEntries.keys())
        matches.sort()
        for match in matches:
            match2workerDump.append("  %s" % match)
            match2entriesDump.append("  %s" % match)
            for worker in self._match2workers(match):
                match2workerDump.append("    %s" % worker)
            for re in self._match2entries(match):
                match2entriesDump.append("    %s" % re)

        dump.append("\n~~~ Match -> Workers ~~~\n%s\n" %
                    "\n".join(match2workerDump))

        dump.append("~~~ Match -> Entries ~~~\n%s\n" %
                    "\n".join(match2entriesDump))

        dump.append("~~~ (source,nlri) ->  entries ~~~")
        for ((source, nlri), entry) in self._source_nlri2entry.iteritems():
            dump.append("  (%s, %s): %s" % (source, nlri, entry))

        log.debug("RouteTableManager data dump:\n\n%s\n", "\n".join(dump))

    # Looking Glass #####

    def getLGMap(self):
        return {"workers": (LGMap.COLLECTION,
                (self.getLGWorkerList, self.getLGWorkerFromPathItem)),
                "routes": (LGMap.SUBTREE, self.getLGRoutes)}

    def getLGRoutes(self, pathPrefix):
        result = {}

        match_IPVPN = Match(
            AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn), Subscription.ANY_RT)
        match_EVPN = Match(
            AFI(AFI.l2vpn), SAFI(SAFI.evpn), Subscription.ANY_RT)
        match_RTC = Match(AFI(AFI.ipv4), SAFI(SAFI.rtc), Subscription.ANY_RT)
        for match in [match_IPVPN, match_EVPN, match_RTC]:
            matchResult = []
            if match in self._match2workersAndEntries:
                for entry in self._match2entries(match):
                    matchResult.append(
                        entry.getLookingGlassInfo(pathPrefix))
            result[repr(match)] = matchResult
        return result

    def getLGWorkerList(self):
        return [{"id": worker.name} for worker in self._worker2matches.keys()]

    def getLGWorkerFromPathItem(self, pathItem):
        # TODO(tmmorin): do a hash-lookup instead of looping the list
        for worker in self._worker2matches.keys():
            if worker.name == pathItem:
                return worker

    def getAllRoutesButRTC(self):
        try:
            return [re for re in
                    self._match2workersAndEntries[Match(Subscription.ANY_AFI,
                                                        Subscription.ANY_SAFI,
                                                        Subscription.ANY_RT)
                                                  ].entries
                    if not (re.afi == AFI(AFI.ipv4) and
                            re.safi == SAFI(SAFI.rtc))
                    ]
        except KeyError:
            return []

    def getLocalRoutesCount(self):
        return reduce(
            lambda count, entry:
            count + (not isinstance(entry.source, BGPPeerWorker)),
            self.getAllRoutesButRTC(),
            0)

    def getReceivedRoutesCount(self):
        return reduce(
            lambda count, entry:
            count + isinstance(entry.source, BGPPeerWorker),
            self.getAllRoutesButRTC(),
            0)



# encoding: utf-8
"""
configuration.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import os
import stat
from pprint import pformat
from copy import deepcopy

from bagpipe.exabgp.structure.ip         import to_IP,to_Route
from bagpipe.exabgp.structure.asn        import ASN
from bagpipe.exabgp.structure.neighbor   import Neighbor
from bagpipe.exabgp.structure.protocol   import NamedProtocol
from bagpipe.exabgp.structure.icmp       import NamedICMPType,NamedICMPCode
from bagpipe.exabgp.structure.tcpflags   import NamedTCPFlags
from bagpipe.exabgp.structure.fragments  import NamedFragments
from bagpipe.exabgp.message.open         import HoldTime,RouterID
#from bagpipe.exabgp.message.update.route import Route
from bagpipe.exabgp.message.update.flow  import BinaryOperator,NumericOperator
from bagpipe.exabgp.message.update.flow  import Flow,Source,Destination,SourcePort,DestinationPort,AnyPort,IPProtocol,TCPFlag,Fragment,PacketLength,ICMPType,ICMPCode,DSCP
from bagpipe.exabgp.message.update.attribute             import AttributeID #,Attribute
from bagpipe.exabgp.message.update.attribute.origin      import Origin
from bagpipe.exabgp.message.update.attribute.nexthop     import NextHop
from bagpipe.exabgp.message.update.attribute.aspath      import ASPath
from bagpipe.exabgp.message.update.attribute.med         import MED
from bagpipe.exabgp.message.update.attribute.localpref   import LocalPreference
from bagpipe.exabgp.message.update.attribute.communities import Community,Communities,ECommunity,ECommunities,to_ExtendedCommunity,to_FlowTrafficRate,to_RouteTargetCommunity_00,to_RouteTargetCommunity_01

from bagpipe.exabgp.log import Logger
logger = Logger()

# Duck class, faking part of the Attribute interface
# We add this to routes when when need o split a route in smaller route
# The value stored is the longer netmask we want to use
# As this is not a real BGP attribute this stays in the configuration file

class Split (int):
	ID = AttributeID.INTERNAL_SPLIT
	MULTIPLE = False


class Watchdog (str):
	ID = AttributeID.INTERNAL_WATCHDOG
	MULTIPLE = False

class Withdrawn (object):
	ID = AttributeID.INTERNAL_WITHDRAWN
	MULTIPLE = False

class Configuration (object):
	TTL_SECURITY = 255
	debug = os.environ.get('RAISE_CONFIGURATION',None) != None

	_str_route_error = \
	'community, extended-communities and as-path can take a single community as parameter.\n' \
	'only next-hop is mandatory\n' \
	'\n' \
	'syntax:\n' \
	'route 10.0.0.1/22 {\n' \
	'  next-hop 192.0.1.254;\n' \
	'  origin IGP|EGP|INCOMPLETE;\n' \
	'  as-path [ ASN1 ASN2 ];\n' \
	'  med 100;\n' \
	'  local-preference 100;\n' \
	'  community [ 65000 65001 65002 ];\n' \
	'  extended-community [ target:1234:5.6.7.8 target:1.2.3.4:5678 origin:1234:5.6.7.8 origin:1.2.3.4:5678 0x0002FDE800000001 ]\n' \
	'  label [ 100 200 ];\n' \
	'  split /24\n' \
	'  watchdog watchog-name\n' \
	'  withdrawn\n' \
	'}\n' \
	'\n' \
	'syntax:\n' \
	'route 10.0.0.1/22 next-hop 192.0.2.1' \
	' origin IGP|EGP|INCOMPLETE' \
	' as-path ASN' \
	' med 100' \
	' local-preference 100' \
	' community 65000' \
	' label 150' \
	' split /24' \
	';\n' \


	_str_flow_error = \
	'syntax: flow {\n' \
	'          match {\n' \
	'             source 10.0.0.0/24;\n' \
	'             destination 10.0.1.0/24;\n' \
	'             port 25;\n' \
	'             source-port >1024\n' \
	'             destination-port =80 =3128 >8080&<8088;\n' \
	'             protocol [ udp tcp ];\n' \
	'             fragment [ not-a-fragment dont-fragment is-fragment first-fragment last-fragment ];\n' \
	'             packet-length >200&<300 >400&<500;'
	'          }\n' \
	'          then {\n' \
	'             discard;\n' \
	'             rate-limit 9600;\n' \
	'             redirect 30740:12345;\n' \
	'             redirect 1.2.3.4:5678;\n' \
	'          }\n' \
	'        }\n\n' \
	'one or more match term, one action\n' \
	'fragment code is totally untested\n' \

	_str_process_error = \
	'syntax: process name-of-process {\n' \
	'          run /path/to/command with its args;\n' \
	'        }\n\n' \

	def __init__ (self,fname,text=False):
		self._text = text
		self._fname = fname
		self._clear()

	def _clear (self):
		self.process = {}
		self.neighbor = {}
		self.error = ''
		self._neighbor = {}
		self._scope = []
		self._location = ['root']
		self._line = []
		self._error = ''
		self._number = 1

	# Public Interface

	def reload (self):
		if self._text:
			self._tokens = self._tokenise(self._fname.split('\n'))
		else:
			try:
				f = open(self._fname,'r')
				self._tokens = self._tokenise(f)
				f.close()
			except IOError,e:
				error = str(e)
				if error.count(']'):
					self.error = error.split(']')[1].strip()
				else:
					self.error = error
				if self.debug: raise
				return False

		self._clear()

		r = False
		while not self.finished():
			r = self._dispatch(self._scope,'configuration',['group','neighbor'],[])
			if r is False: break

		if r in [True,None]:
			self.neighbor = self._neighbor
			return True

		self.error = "\nsyntax error in section %s\nline %d : %s\n\n%s" % (self._location[-1],self.number(),self.line(),self._error)
		return False

	def parse_single_route (self,command):
		tokens = command.split(' ')[1:]
		if len(tokens) < 4:
			return False
		if tokens[0] != 'route':
			return False
		scope = [{}]
		if not self._single_static_route(scope,tokens[1:]):
			return None
		return scope[0]['routes'][0]

	def parse_single_flow (self,command):
		self._tokens = self._tokenise(' '.join(command.split(' ')[2:]).split('\\n'))
		scope = [{}]
		if not self._dispatch(scope,'flow',['route',],[]):
			return None
		if not self._check_flow_route(scope):
			return None
		return scope[0]['routes'][0]

	def add_route_all_peers (self,route):
		for neighbor in self.neighbor:
			self.neighbor[neighbor].add_route(route)

	def remove_route_all_peers (self,route):
		result = False
		for neighbor in self.neighbor:
			if self.neighbor[neighbor].remove_route(route):
				result = True
		return result

	# Tokenisation

	def _tokenise (self,text):
		r = []
		config = ''
		for line in text:
			replaced = line.strip().replace('\t',' ').replace(']',' ]').replace('[','[ ').lower()
			config += line
			if not replaced:
				continue
			if replaced.startswith('#'):
				continue
			if replaced[:3] == 'md5':
				password = line.strip()[3:].strip()
				if password[-1] == ';':
					password = password[:-1]
				r.append(['md5',password,';'])
			else:
				r.append([t for t in replaced[:-1].split(' ') if t] + [replaced[-1]])
		logger.config(config)
		return r

	def tokens (self):
		self._number += 1
		self._line = self._tokens.pop(0)
		return self._line

	def number (self):
		return self._number

	def line (self):
		return ' '.join(self._line)

	def finished (self):
		return len(self._tokens) == 0

	# Flow control ......................

	# name is not used yet but will come really handy if we have name collision :D
	def _dispatch (self,scope,name,multi,single):
		try:
			tokens = self.tokens()
		except IndexError:
			self._error = 'configuration file incomplete (most likely missing })'
			if self.debug: raise
			return False
		logger.configuration('analysing tokens %s ' % str(tokens))
		logger.configuration('  valid block options %s' % str(multi))
		logger.configuration('  valid parameters    %s' % str(single))
		end = tokens[-1]
		if multi and end == '{':
			self._location.append(tokens[0])
			return self._multi_line(scope,name,tokens[:-1],multi)
		if single and end == ';':
			return self._single_line(scope,name,tokens[:-1],single)
		if end == '}':
			if len(self._location) == 1:
				self._error = 'closing too many parenthesis'
				return False
			self._location.pop(-1)
			return None
		return False

	def _multi_line (self,scope,name,tokens,valid):
		command = tokens[0]
		if valid and command not in valid:
			self._error = 'option %s in not valid here' % command
			return False

		if name == 'configuration':
			if  command == 'neighbor':
				if self._multi_neighbor(scope,tokens[1:]):
					return self._make_neighbor(scope)
				return False
			if  command == 'group':
				if len(tokens) != 2:
					self._error = 'syntax: group <name> { <options> }'
					return False
				return self._multi_group(scope,tokens[1])

		if name == 'group':
			if  command == 'neighbor':
				if self._multi_neighbor(scope,tokens[1:]):
					return self._make_neighbor(scope)
				return False
			if command == 'static': return self._multi_static(scope,tokens[1:])
			if command == 'flow': return self._multi_flow(scope,tokens[1:])
			if command == 'process': return self._multi_process(scope,tokens[1:])

		if name == 'neighbor':
			if command == 'static': return self._multi_static(scope,tokens[1:])
			if command == 'flow': return self._multi_flow(scope,tokens[1:])
			if command == 'process': return self._multi_process(scope,tokens[1:])

		if name == 'static':
			if command == 'route':
				if self._multi_static_route(scope,tokens[1:]):
					return self._check_static_route(scope)
				return False

		if name == 'flow':
			if command == 'route':
				if self._multi_flow_route(scope,tokens[1:]):
					return self._check_flow_route(scope)
				return False

		if name == 'flow-route':
			if command == 'match':
				if self._multi_match(scope,tokens[1:]):
					return True
				return False
			if command == 'then':
				if self._multi_then(scope,tokens[1:]):
					return True
				return False
		return False

	def _single_line (self,scope,name,tokens,valid):
		command = tokens[0]
		if valid and command not in valid:
			self._error = 'invalid keyword "%s"' % command
			return False

		if command == 'description': return self._set_description(scope,tokens[1:])
		if command == 'router-id': return self._set_router_id(scope,'router-id',tokens[1:])
		if command == 'local-address': return self._set_ip(scope,'local-address',tokens[1:])
		if command == 'local-as': return self._set_asn(scope,'local-as',tokens[1:])
		if command == 'peer-as': return self._set_asn(scope,'peer-as',tokens[1:])
		if command == 'hold-time': return self._set_holdtime(scope,'hold-time',tokens[1:])
		if command == 'graceful-restart': return self._set_gracefulrestart(scope,'graceful-restart',tokens[1:])
		if command == 'md5': return self._set_md5(scope,'md5',tokens[1:])
		if command == 'ttl-security': return self._set_ttl(scope,'ttl-security',tokens[1:])
		if command == 'multi-session': return self._set_multisession(scope,'multi-session',tokens[1:])

		if command == 'route': return self._single_static_route(scope,tokens[1:])
		if command == 'origin': return self._route_origin(scope,tokens[1:])
		if command == 'as-path': return self._route_aspath(scope,tokens[1:])
		if command == 'as-sequence': return self._route_aspath(scope,tokens[1:])
		if command == 'med': return self._route_med(scope,tokens[1:])
		if command == 'next-hop': return self._route_next_hop(scope,tokens[1:])
		if command == 'local-preference': return self._route_local_preference(scope,tokens[1:])
		if command == 'community': return self._route_community(scope,tokens[1:])
		if command == 'split': return self._route_split(scope,tokens[1:])
		if command == 'label': return self._route_label(scope,tokens[1:])
		if command == 'watchdog': return self._route_watchdog(scope,tokens[1:])
		if command == 'withdrawn': return self._route_withdrawn(scope,tokens[1:])

		if command == 'source': return self._flow_source(scope,tokens[1:])
		if command == 'destination': return self._flow_destination(scope,tokens[1:])
		if command == 'port': return self._flow_route_anyport(scope,tokens[1:])
		if command == 'source-port': return self._flow_route_source_port(scope,tokens[1:])
		if command == 'destination-port': return self._flow_route_destination_port(scope,tokens[1:])
		if command == 'protocol': return self._flow_route_protocol(scope,tokens[1:])
		if command == 'tcp-flags': return self._flow_route_tcp_flags(scope,tokens[1:])
		if command == 'icmp-type': return self._flow_route_icmp_type(scope,tokens[1:])
		if command == 'icmp-code': return self._flow_route_icmp_code(scope,tokens[1:])
		if command == 'fragment': return self._flow_route_fragment(scope,tokens[1:])
		if command == 'dscp': return self._flow_route_dscp(scope,tokens[1:])
		if command == 'packet-length': return self._flow_route_packet_length(scope,tokens[1:])
		if command == 'discard': return self._flow_route_discard(scope,tokens[1:])
		if command == 'rate-limit': return self._flow_route_rate_limit(scope,tokens[1:])
		if command == 'redirect': return self._flow_route_redirect(scope,tokens[1:])

		if command == 'run': return self._set_process_run(scope,'process-run',tokens[1:])
		if command == 'parse-routes': return self._set_process_parse_routes(scope,'parse-routes',tokens[1:])

		return False

	# Group Watchdog

	def _multi_process (self,scope,tokens):
		if len(tokens) != 1:
			self._error = self._str_process_error
			return False
		while True:
			r = self._dispatch(scope,'process',[],['run','parse-routes'])
			if r is False: return False
			if r is None: break
		self.process.setdefault(tokens[0],{})['run'] = scope[-1].pop('process-run')
		self.process[tokens[0]]['receive-routes'] = scope[-1].get('parse-routes',False)
		if 'peer-address' in scope[-1]:
			self.process[tokens[0]]['neighbor'] = scope[-1]['peer-address']
		else:
			self.process[tokens[0]]['neighbor'] = '*'
		return True

	def _set_process_parse_routes (self,scope,command,value):
		scope[-1][command] = True
		return True

	def _set_process_run (self,scope,command,value):
		line = ' '.join(value).strip()
		if len(line) > 2 and line[0] == line[-1] and line[0] in ['"',"'"]:
			line = line[1:-1]
		if ' ' in line:
			prg,args = line.split(' ',1)
		else:
			prg = line
			args = ''

		if not prg:
			self._error = 'prg requires the program to prg as an argument (quoted or unquoted)'
			if self.debug: raise
			return False
		if prg[0] != '/':
			if prg.startswith('etc/exabgp'):
				parts = prg.split('/')
				path = [os.environ.get('ETC','etc'),] + parts[2:]
				prg = os.path.join(*path)
			else:
				prg = os.path.join(os.getcwd(),prg)
		if not os.path.exists(prg):
			self._error = 'can not locate the the program "%s"' % prg
			if self.debug: raise
			return False

		# XXX: Yep, race conditions are possible, those are sanity checks not security ones ...
		s = os.stat(prg)

		if stat.S_ISDIR(s.st_mode):
			self._error = 'can not execute directories "%s"' % prg
			if self.debug: raise
			return False

		if s.st_mode & stat.S_ISUID:
			self._error = 'refusing to run setuid programs "%s"' % prg
			if self.debug: raise
			return False

		check = stat.S_IXOTH
		if s.st_uid == os.getuid():
			check |= stat.S_IXUSR
		if s.st_gid == os.getgid():
			check |= stat.S_IXGRP

		if not check & s.st_mode:
			self._error = 'exabgp will not be able to run this program "%s"' % prg
			if self.debug: raise
			return False

		if args:
			scope[-1][command] = [prg] + args.split(' ')
		else:
			scope[-1][command] = [prg,]
		return True

	def _route_watchdog (self,scope,tokens):
		try:
			scope[-1]['routes'][-1].attributes.add(Watchdog(tokens.pop(0)))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _route_withdrawn (self,scope,tokens):
		try:
			scope[-1]['routes'][-1].attributes.add(Withdrawn())
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	# Group Neighbor

	def _multi_group (self,scope,address):
		scope.append({})
		while True:
			r = self._dispatch(scope,'group',['static','flow','neighbor','process'],['description','router-id','local-address','local-as','peer-as','hold-time','graceful-restart','md5','ttl-security','multi-session'])
			if r is False:
				return False
			if r is None:
				scope.pop(-1)
				return True

	def _make_neighbor (self,scope):
		# we have local_scope[-2] as the group template and local_scope[-1] as the peer specific
		if len(scope) > 1:
			for key,content in scope[-2].iteritems():
				if key not in scope[-1]:
					scope[-1][key] = deepcopy(content)

		logger.configuration("\nPeer configuration complete :")
		for _key in scope[-1].keys():
			for _line in pformat(scope[-1][_key],3,3,3).split('\n'):
				logger.configuration("   %s: %s" %(_key,_line))
		logger.configuration("\n")

		neighbor = Neighbor()
		for local_scope in scope:
			v = local_scope.get('router-id','')
			if v: neighbor.router_id = v
			v = local_scope.get('peer-address','')
			if v: neighbor.peer_address = v
			v = local_scope.get('local-address','')
			if v: neighbor.local_address = v
			v = local_scope.get('local-as','')
			if v: neighbor.local_as = v
			v = local_scope.get('peer-as','')
			if v: neighbor.peer_as = v
			v = local_scope.get('hold-time','')
			if v: neighbor.hold_time = v
			neighbor.parse_routes = local_scope.get('parse-routes',False)
			v = local_scope.get('routes',[])
			for route in v:
				# This add the family to neighbor.families()
				neighbor.add_route(route)

		# drop the neighbor
		local_scope = scope.pop(-1)
		neighbor.description = local_scope.get('description','')

		neighbor.graceful_restart = local_scope.get('graceful-restart',0)
		if neighbor.graceful_restart is None:
			# README: Should it be a subclass of int ?
			neighbor.graceful_restart = int(neighbor.hold_time)

		neighbor.md5 = local_scope.get('md5',None)
		neighbor.ttl = local_scope.get('ttl-security',None)
		neighbor.multisession = local_scope.get('multi-session',False)

		missing = neighbor.missing()
		if missing:
			self._error = 'incomplete neighbor, missing %s' % missing
			return False
		if neighbor.local_address.afi != neighbor.peer_address.afi:
			self._error = 'local-address and peer-address must be of the same family'
			return False
		if self._neighbor.has_key(neighbor.peer_address.ip):
			self._error = 'duplicate peer definition %s' % neighbor.peer_address.ip
			return False

		if neighbor.multisession:
			for family in neighbor.families():
				afi,safi = family
				family_n = deepcopy(neighbor)
				for f in neighbor.families():
					if f == family:
						continue
					family_n.remove_family(f)
				self._neighbor[family_n.name()] = family_n
		else:
			self._neighbor[neighbor.name()] = neighbor
		return True


	def _multi_neighbor (self,scope,tokens):
		if len(tokens) != 1:
			self._error = 'syntax: neighbor <ip> { <options> }'
			return False

		address = tokens[0]
		scope.append({})
		try:
			scope[-1]['peer-address'] = to_IP(address)
		except:
			self._error = '"%s" is not a valid IP address' % address
			if self.debug: raise
			return False
		while True:
		 	r = self._dispatch(scope,'neighbor',['static','flow','process'],['description','router-id','local-address','local-as','peer-as','hold-time','graceful-restart','md5','ttl-security','multi-session'])
			if r is False: return False
			if r is None: return True

	# Command Neighbor

	def _set_router_id (self,scope,command,value):
		try:
			ip = RouterID(value[0])
		except (IndexError,ValueError):
			self._error = '"%s" is an invalid IP address' % ' '.join(value)
			if self.debug: raise
			return False
		scope[-1][command] = ip
		return True

	def _set_description (self,scope,tokens):
		text = ' '.join(tokens)
		if len(text) < 2 or text[0] != '"' or text[-1] != '"' or text[1:-1].count('"'):
			self._error = 'syntax: description "<description>"'
			return False
		scope[-1]['description'] = text[1:-1]
		return True

	# will raise ValueError if the ASN is not correct
	def _newASN (self,value):
		if value.count('.'):
			high,low = value.split('.',1)
			asn = (int(high) << 16) + int(low)
		else:
			asn = int(value)
		return ASN(asn)

	def _set_asn (self,scope,command,value):
		try:
			scope[-1][command] = self._newASN(value[0])
			return True
		except ValueError:
			self._error = '"%s" is an invalid ASN' % ' '.join(value)
			if self.debug: raise
			return False

	def _set_ip (self,scope,command,value):
		try:
			ip = to_IP(value[0])
		except (IndexError,ValueError):
			self._error = '"%s" is an invalid IP address' % ' '.join(value)
			if self.debug: raise
			return False
		scope[-1][command] = ip
		return True

	def _set_holdtime (self,scope,command,value):
		try:
			holdtime = HoldTime(value[0])
			if holdtime < 0:
				raise ValueError('holdtime can not be negative')
			if holdtime >= pow(2,16):
				raise ValueError('holdtime must be smaller than %d' % pow(2,16))
			scope[-1][command] = holdtime
			return True
		except ValueError:
			self._error = '"%s" is an invalid hold-time' % ' '.join(value)
			if self.debug: raise
			return False

	def _set_gracefulrestart (self,scope,command,value):
		if not len(value):
			scope[-1][command] = None
			return True
		try:
			# README: Should it be a subclass of int ?
			grace = int(value[0])
			if grace < 0:
				raise ValueError('graceful-restart can not be negative')
			if grace >= pow(2,16):
				raise ValueError('graceful-restart must be smaller than %d' % pow(2,16))
			scope[-1][command] = grace
			return True
		except ValueError:
			self._error = '"%s" is an invalid graceful-restart time' % ' '.join(value)
			if self.debug: raise
			return False
		return True

	def _set_md5 (self,scope,command,value):
		md5 = value[0]
		if len(md5) > 2 and md5[0] == md5[-1] and md5[0] in ['"',"'"]:
			md5 = md5[1:-1]
		if len(md5) > 80:
			self._error = 'md5 password must be no larger than 80 characters'
			if self.debug: raise
			return False
		if not md5:
			self._error = 'md5 requires the md5 password as an argument (quoted or unquoted)'
			if self.debug: raise
			return False
		scope[-1][command] = md5
		return True

	def _set_ttl (self,scope,command,value):
		if not len(value):
			scope[-1][command] = self.TTL_SECURITY
			return True
		try:
			# README: Should it be a subclass of int ?
			ttl = int(value[0])
			if ttl < 0:
				raise ValueError('ttl-security can not be negative')
			if ttl >= 255:
				raise ValueError('ttl must be smaller than 256')
			scope[-1][command] = ttl
			return True
		except ValueError:
			self._error = '"%s" is an invalid ttl-security' % ' '.join(value)
			if self.debug: raise
			return False
		return True

	def _set_multisession (self,scope,command,value):
		scope[-1][command] = True
		return True

	#  Group Static ................

	def _multi_static (self,scope,tokens):
		if len(tokens) != 0:
			self._error = 'syntax: static { route; route; ... }'
			return False
		while True:
		 	r = self._dispatch(scope,'static',['route',],['route',])
			if r is False: return False
			if r is None: return True

	# Group Route  ........

	def _split_last_route (self,scope):
		# if the route does not need to be broken in smaller routes, return
		route = scope[-1]['routes'][-1]
		if not AttributeID.INTERNAL_SPLIT in route.attributes:
			return True

		# ignore if the request is for an aggregate, or the same size
		mask = route.nlri.mask
		split = route.attributes[AttributeID.INTERNAL_SPLIT]
		if mask >= split:
			return True

		# remove the route, we are going to replace it
		route = scope[-1]['routes'].pop(-1)

		# calculate the number of IP in the /<size> of the new route
		increment = pow(2,(len(route.nlri)*8) - split)
		# how many new routes are we going to create from the initial one
		number = pow(2,split - route.nlri.mask)

		# convert the IP into a integer/long
		ip = 0
		for c in route.nlri.raw:
			ip = ip << 8
			ip += ord(c)

		# route is becoming a template we will clone (deepcopy) so change its netmask
		route.nlri.mask = split

		# generate the new routes
		for _ in range(number):
			r = deepcopy(route)
			# convert the ip to a network packed format
			ipn = ip
			i = ''
			while ipn:
				lower = ipn&0xFF
				ipn = ipn >> 8
				i = chr(lower) + i

			# change the route network
			r.nlri.update_raw(i)
			# update ip to the next route
			ip += increment

			# save route
			scope[-1]['routes'].append(r)

		# route is no longer needed - delete it explicitely
		del(route)
		return True

	def _insert_static_route (self,scope,tokens):
		try:
			ip = tokens.pop(0)
		except IndexError:
			self._error = self._str_route_error
			if self.debug: raise
			return False
		try:
			ip,nm = ip.split('/')
		except ValueError:
			nm = '32'
		try:
			route = to_Route(ip,nm)
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

		if not scope[-1].has_key('routes'):
			scope[-1]['routes'] = []

		scope[-1]['routes'].append(route)
		return True

	def pop_last_static_route (self,scope):
		route = scope[-1]['routes'][-1]
		scope[-1]['routes'] = scope[-1]['routes'][:-1]
		return route

	def remove_route (self,route,scope):
		for r in scope[-1]['routes']:
			if r == route:
				scope[-1]['routes'].remove(r)
				return True
		return False

	def _check_static_route (self,scope):
		route = scope[-1]['routes'][-1]
		if not route.attributes.has(AttributeID.NEXT_HOP):
			self._error = 'syntax: route IP/MASK { next-hop IP; }'
			return False
		return True

	def _multi_static_route (self,scope,tokens):
		if len(tokens) != 1:
			self._error = self._str_route_error
			return False

		if not self._insert_static_route(scope,tokens):
			return False

		while True:
			r = self._dispatch(scope,'route',[],['next-hop','origin','as-path','as-sequence','med','local-preference','community','extended-community','split','label','watchdog','withdrawn'])
			if r is False: return False
			if r is None: return self._split_last_route(scope)

	def _single_static_route (self,scope,tokens):
		if len(tokens) <3:
			return False

		if not self._insert_static_route(scope,tokens):
			return False

		if tokens.pop(0) != 'next-hop':
			return False

		if not self._route_next_hop(scope,tokens):
			return False

		while len(tokens):
			command = tokens.pop(0)
			if command == 'withdrawn':
				if self._route_withdrawn(scope,tokens):
					continue
				return False

			if len(tokens) < 1:
				return False

			if command == 'origin':
				if self._route_origin(scope,tokens):
					continue
				return False
			if command == 'as-path':
				if self._route_aspath(scope,tokens):
					continue
				return False
			if command == 'as-sequence':
				if self._route_aspath(scope,tokens):
					continue
				return False
			if command == 'med':
				if self._route_med(scope,tokens):
					continue
				return False
			if command == 'local-preference':
				if self._route_local_preference(scope,tokens):
					continue
				return False
			if command == 'community':
				if self._route_community(scope,tokens):
					continue
				return False
			if command == 'extended-community':
				if self._route_extended_community(scope,tokens):
					continue
				return False
			if command == 'split':
				if self._route_split(scope,tokens):
					continue
				return False
			if command == 'label':
				if self._route_label(scope,tokens):
					continue
				return False
			if command == 'watchdog':
				if self._route_watchdog(scope,tokens):
					continue
				return False
			return False
		return self._split_last_route(scope)

	# Command Route

	def _route_next_hop (self,scope,tokens):
		try:
			scope[-1]['routes'][-1].attributes.add(NextHop(to_IP(tokens.pop(0))))
			return True
		except:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _route_origin (self,scope,tokens):
		data = tokens.pop(0).lower()
		if data == 'igp':
			scope[-1]['routes'][-1].attributes.add(Origin(0x00))
			return True
		if data == 'egp':
			scope[-1]['routes'][-1].attributes.add(Origin(0x01))
			return True
		if data == 'incomplete':
			scope[-1]['routes'][-1].attributes.add(Origin(0x02))
			return True
		self._error = self._str_route_error
		if self.debug: raise
		return False

	def _route_aspath (self,scope,tokens):
		aspath = ASPath()
		asn = tokens.pop(0)
		try:
			if asn == '[':
				while True:
					try:
						asn = tokens.pop(0)
					except IndexError:
						self._error = self._str_route_error
						if self.debug: raise
						return False
					if asn == ']':
						break
					aspath.add(self._newASN(asn))
			else:
				aspath.add(self._newASN(asn))
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False
		scope[-1]['routes'][-1].attributes.add(aspath)
		return True

	def _route_med (self,scope,tokens):
		try:
			scope[-1]['routes'][-1].attributes.add(MED(int(tokens.pop(0))))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _route_local_preference (self,scope,tokens):
		try:
			scope[-1]['routes'][-1].attributes.add(LocalPreference(int(tokens.pop(0))))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _parse_community (self,scope,data):
		separator = data.find(':')
		if separator > 0:
			prefix = int(data[:separator])
			suffix = int(data[separator+1:])
			if prefix >= pow(2,16):
				raise ValueError('invalid community %s (prefix too large)' % data)
			if suffix >= pow(2,16):
				raise ValueError('invalid community %s (suffix too large)' % data)
			return Community((prefix<<16) + suffix)
		elif len(data) >=2 and data[1] in 'xX':
			value = long(data,16)
			if value >= pow(2,32):
				raise ValueError('invalid community %s (too large)' % data)
			return Community(value)
		else:
			low = data.lower()
			if low == 'no-export':
				data = 0xFFFFFF01
			elif low == 'no-advertise':
				data = 0xFFFFFF02
			elif low == 'no-export-subconfed':
				data = 0xFFFFFF03
			value = long(data)
			if value >= pow(2,32):
				raise ValueError('invalid community %s (too large)' % data)
			return Community(value)

	def _route_community (self,scope,tokens):
		communities = Communities()
		community = tokens.pop(0)
		try:
			if community == '[':
				while True:
					try:
						community = tokens.pop(0)
					except IndexError:
						self._error = self._str_route_error
						if self.debug: raise
						return False
					if community == ']':
						break
					communities.add(self._parse_community(scope,community))
			else:
				communities.add(self._parse_community(scope,community))
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False
		scope[-1]['routes'][-1].attributes.add(communities)
		return True

	def _parse_extended_community (self,scope,data):
		if data[:2].lower() == '0x':
			try:
				raw = ''
				for i in range(2,len(data),2):
					raw += chr(int(data[i:i+2],16))
			except ValueError:
				raise ValueError('invalid extended community %s' % data)
			if len(raw) != 8:
				raise ValueError('invalid extended community %s' % data)
			return ECommunity(raw)
		elif data.count(':'):
			return to_ExtendedCommunity(data)
		else:
			raise ValueError('invalid extended community %s - lc+gc' % data)

	def _route_extended_community (self,scope,tokens):
		extended_communities = ECommunities()
		extended_community = tokens.pop(0)
		try:
			if extended_community == '[':
				while True:
					try:
						extended_community = tokens.pop(0)
					except IndexError:
						self._error = self._str_route_error
						if self.debug: raise
						return False
					if extended_community == ']':
						break
					extended_communities.add(self._parse_extended_community(scope,extended_community))
			else:
				extended_communities.add(self._parse_extended_community(scope,extended_community))
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False
		scope[-1]['routes'][-1].attributes.add(extended_communities)
		return True


	def _route_split (self,scope,tokens):
		try:
			size = tokens.pop(0)
			if not size or size[0] != '/':
				raise ValueError('route "as" require a CIDR')
			scope[-1]['routes'][-1].attributes.add(Split(int(size[1:])))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _route_label (self,scope,tokens):
		communities = Communities()
		community = tokens.pop(0)
		try:
			if community == '[':
				while True:
					try:
						community = tokens.pop(0)
					except IndexError:
						self._error = self._str_route_error
						if self.debug: raise
						return False
					if community == ']':
						break
					communities.add(self._parse_community(community))
			else:
				communities.add(self._parse_community(community))
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False
		scope[-1]['routes'][-1].attributes.add(communities)
		return True

	# Group Flow  ........

	def _multi_flow (self,scope,tokens):
		if len(tokens) != 0:
			self._error = self._str_flow_error
			return False

		while True:
			r = self._dispatch(scope,'flow',['route',],[])
			if r is False: return False
			if r is None: break
		return True

	def _insert_flow_route (self,scope,tokens=None):
		try:
			flow = Flow()
		except ValueError:
			self._error = self._str_flow_error
			if self.debug: raise
			return False

		if not scope[-1].has_key('routes'):
			scope[-1]['routes'] = []

		scope[-1]['routes'].append(flow)
		return True

	def _check_flow_route (self,scope):
		logger.configuration('warning: no check on flows are implemented')
		return True

	def _multi_flow_route (self,scope,tokens):
		if len(tokens) > 1:
			self._error = self._str_flow_error
			return False

		if not self._insert_flow_route(scope):
			return False

		r = self._dispatch(scope,'flow-route',['match',],[])
		if r is False: return False
		r = self._dispatch(scope,'flow-route',['then',],[])
		return r

	# ..........................................

	def _multi_match (self,scope,tokens):
		if len(tokens) != 0:
			self._error = self._str_flow_error
			return False

		while True:
			r = self._dispatch(scope,'flow-match',[],['source','destination','port','source-port','destination-port','protocol','tcp-flags','icmp-type','icmp-code','fragment','dscp','packet-length'])
			if r is False: return False
			if r is None: break
		return True

	def _multi_then (self,scope,tokens):
		if len(tokens) != 0:
			self._error = self._str_flow_error
			return False

		while True:
			r = self._dispatch(scope,'flow-then',[],['discard','rate-limit','redirect','community'])
			if r is False: return False
			if r is None: break
		return True

	# Command Flow

	def _flow_source (self,scope,tokens):
		try:
			ip,nm = tokens.pop(0).split('/')
			scope[-1]['routes'][-1].add_and(Source(ip,nm))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _flow_destination (self,scope,tokens):
		try:
			ip,nm = tokens.pop(0).split('/')
			scope[-1]['routes'][-1].add_and(Destination(ip,nm))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	# to parse the port configuration of flow

	def _operator (self,string):
		try:
			if string[0] == '=':
				return NumericOperator.EQ,string[1:]
			elif string[0] == '>':
				operator = NumericOperator.GT
			elif string[0] == '<':
				operator = NumericOperator.LT
			else:
				raise ValueError('Invalid operator in test %s' % string)
			if string[1] == '=':
				operator += NumericOperator.EQ
				return operator,string[2:]
			else:
				return operator,string[1:]
		except IndexError:
			raise('Invalid expression (too short) %s' % string)

	def _value (self,string):
		l = 0
		for c in string:
			if c not in ['&',]:
				l += 1
				continue
			break
		return string[:l],string[l:]

	# parse =80 or >80 or <25 or &>10<20
	def _flow_generic_expression (self,scope,tokens,converter,klass):
		try:
			for test in tokens:
				AND = BinaryOperator.NOP
				while test:
					operator,_ = self._operator(test)
					value,test = self._value(_)
					try:
						number = int(value)
					except ValueError:
						number = converter(value)
					scope[-1]['routes'][-1].add_or(klass(AND|operator,number))
					if test:
						if test[0] == '&':
							AND = BinaryOperator.AND
							test = test[1:]
							if not test:
								raise ValueError("Can not finish an expresion on an &")
						else:
							raise ValueError("Unknown binary operator %s" % test[0])
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	# parse [ content1 content2 content3 ]
	def _flow_generic_list (self,scope,tokens,converter,klass):
		name = tokens.pop(0)
		AND = BinaryOperator.NOP
		try:
			if name == '[':
				while True:
					name = tokens.pop(0)
					if name == ']':
						break
					try:
						try:
							number = int(name)
						except ValueError:
							number = converter(name)
						scope[-1]['routes'][-1].add_or(klass(NumericOperator.EQ|AND,number))
					except IndexError:
						self._error = self._str_flow_error
						if self.debug: raise
						return False
			else:
				try:
					number = int(name)
				except ValueError:
					number = converter(name)
				scope[-1]['routes'][-1].add_or(klass(NumericOperator.EQ|AND,number))
		except ValueError:
			self._error = self._str_flow_error
			if self.debug: raise
			return False
		return True

	def _flow_generic_condition (self,scope,tokens,converter,klass):
		if tokens[0][0] in ['=','>','<']:
			return self._flow_generic_expression(scope,tokens,converter,klass)
		return self._flow_generic_list(scope,tokens,converter,klass)

	def _flow_route_anyport (self,scope,tokens):
		return self._flow_generic_condition(scope,tokens,int,AnyPort)

	def _flow_route_source_port (self,scope,tokens):
		return self._flow_generic_condition(scope,tokens,int,SourcePort)

	def _flow_route_destination_port (self,scope,tokens):
		return self._flow_generic_condition(scope,tokens,int,DestinationPort)

	def _flow_route_packet_length (self,scope,tokens):
		return self._flow_generic_condition(scope,tokens,int,PacketLength)

	def _flow_route_tcp_flags (self,scope,tokens):
		return self._flow_generic_list(scope,tokens,NamedTCPFlags,TCPFlag)

	def _flow_route_protocol (self,scope,tokens):
		return self._flow_generic_list(scope,tokens,NamedProtocol,IPProtocol)

	def _flow_route_icmp_type (self,scope,tokens):
		return self._flow_generic_list(scope,tokens,NamedICMPType,ICMPType)

	def _flow_route_icmp_code (self,scope,tokens):
		return self._flow_generic_list(scope,tokens,NamedICMPCode,ICMPCode)

	def _flow_route_fragment (self,scope,tokens):
		return self._flow_generic_list(scope,tokens,NamedFragments,Fragment)

	def _flow_route_dscp (self,scope,tokens):
		return self._flow_generic_condition(scope,tokens,int,DSCP)

	def _flow_route_discard (self,scope,tokens):
		# README: We are setting the ASN as zero as that what Juniper (and Arbor) did when we created a local flow route
		try:
			scope[-1]['routes'][-1].add_action(to_FlowTrafficRate(ASN(0),0))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _flow_route_rate_limit (self,scope,tokens):
		# README: We are setting the ASN as zero as that what Juniper (and Arbor) did when we created a local flow route
		try:
			speed = int(tokens[0])
			if speed < 9600 and speed != 0:
				logger.warning("rate-limiting flow under 9600 bytes per seconds may not work","configuration")
			if speed > 1000000000000:
				speed = 1000000000000
				logger.warning("rate-limiting changed for 1 000 000 000 000 bytes from %s" % tokens[0],"configuration")
			scope[-1]['routes'][-1].add_action(to_FlowTrafficRate(ASN(0),speed))
			return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False

	def _flow_route_redirect (self,scope,tokens):
		# README: We are setting the ASN as zero as that what Juniper (and Arbor) did when we created a local flow route
		try:
			prefix,suffix=tokens[0].split(':',1)
			if prefix.count('.'):
				ip = prefix.split('.')
				if len(ip) != 4:
					raise ValueError('')
				ipn = 0
				while ip:
					ipn <<= 8
					ipn += int(ip.pop(0))
				number = int(suffix)
				scope[-1]['routes'][-1].add_action(to_RouteTargetCommunity_01(ipn,number))
				return True
			else:
				asn = int(prefix)
				route_target = int(suffix)
				if asn >= pow(2,16):
					asn = asn & 0xFFFF
				if route_target >= pow(2,32):
					raise ValueError('route target is a 32 bits number, value too large %s' % route_target)
				scope[-1]['routes'][-1].add_action(to_RouteTargetCommunity_00(asn,route_target))
				return True
		except ValueError:
			self._error = self._str_route_error
			if self.debug: raise
			return False




# encoding: utf-8
"""
labels.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from struct import pack

#from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== Community

class Label (object):
	def __init__ (self,label):
		self.label = label

	def pack (self):
		return pack('!L',self.label)

	def __str__ (self):
		return "NOT DONE"

	def __len__ (self):
		return 4

	def __cmp__(self,other):
		if ( not isinstance(other,Label)
			or (self.label != other.label)
			):
			return -1
		else:
			return 0


#!/usr/bin/env python
#
# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import stdout

import os
import urllib2
import json

from optparse import OptionParser

BAGPIPE_PORT = 8082
LOOKING_GLASS_BASE = "looking-glass"

INDENT_INCREMENT = 2


def pretty_print_recurse(data, indent, recursiveRequests, url,
                         alreadyANewLine=False):
    """
    key has already been output, this function will print data and finish at
    a start of line

    returns True if the key output was spread across multiple lines.
    """
    if isinstance(data, dict):
        more = False

        if ("id" in data and "href" in data):
            stdout.write(data["id"])
            del data["id"]

        if ("href" in data):
            more = True

        if more:
            targetURL = data["href"]
            if recursiveRequests:
                if targetURL.startswith(url):
                    response = urllib2.urlopen(targetURL)
                    if response.getcode() == 200:
                        pretty_print_recurse(json.load(response),
                                             indent + INDENT_INCREMENT,
                                             recursiveRequests, targetURL,
                                             alreadyANewLine=False)
                        return True
                    else:
                        stdout.write(" (ERROR %d)", response.getcode())
                        return False

            del data["href"]
            stdout.write(" (...)")
            alreadyANewLine = False

        if len(data) > 0:
            if not alreadyANewLine:
                stdout.write("\n")
            firstVal = True
            for (key, value) in data.iteritems():
                if not firstVal or not alreadyANewLine:
                    stdout.write("%s" % (" " * indent))
                if firstVal:
                    firstVal = False
                stdout.write("%s: " % key)
                pretty_print_recurse(value, indent + INDENT_INCREMENT,
                                     recursiveRequests, url)
        else:
            if more:
                stdout.write("\n")
            else:
                stdout.write("-\n")

    elif isinstance(data, list):

        if len(data) > 0:
            if not alreadyANewLine:
                stdout.write("\n")
                alreadyANewLine = True

            for i in range(len(data)):
                value = data[i]
                stdout.write("%s* " % (" " * indent))
                if isinstance(value, dict) or isinstance(value, list):
                    pretty_print_recurse(value, indent + INDENT_INCREMENT,
                                         recursiveRequests, url,
                                         alreadyANewLine)
                else:
                    stdout.write("%s\n" % value)
                alreadyANewLine = True
        else:
            stdout.write("-\n")

    else:
        if isinstance(data, str) and "\n" in data:
            data = data.strip("\n").replace("\n", "\n%s" % (" " * indent))
            stdout.write("\n%s" % (" " * indent))

        stdout.write("%s\n" % data)
        return False


def main():
    usage = """ %prog [--server <ip>] path to object in looking-glass

e.g.: %prog vpns instances"""
    parser = OptionParser(usage)

    parser.add_option(
        "--server", dest="server", default="127.0.0.1",
        help="IP address of BaGPipe BGP (optional, default: %default)")

    parser.add_option(
        "--port", dest="port", type="int", default=BAGPIPE_PORT,
        help="Port of BaGPipe BGP (optional, default: %default)")

    parser.add_option(
        "--prefix", dest="prefix", default=LOOKING_GLASS_BASE,
        help="Looking-glass URL Prefix (optional, default: %default)")

    parser.add_option(
        "-r", "--recurse", dest="recurse", action="store_true", default=False,
        help="Recurse down into the whole looking-glass (disabled by default)")

    (options, args) = parser.parse_args()

    quoted_args = [urllib2.quote(arg) for arg in args]
    target_url = "http://%s:%d/%s/%s" % (options.server, options.port,
                                         options.prefix, "/".join(quoted_args))
    try:
        os.environ['NO_PROXY'] = options.server
        response = urllib2.urlopen(target_url)

        if response.getcode() == 200:
            data = json.load(response)

            if (isinstance(data, dict) and "href" in data):
                target_url_bis = data["href"]
                response_bis = urllib2.urlopen(target_url_bis)
                if response.getcode() == 200:
                    target_url = target_url_bis
                    data = json.load(response_bis)

            pretty_print_recurse(data, 0, options.recurse, target_url,
                                 alreadyANewLine=True)

    except urllib2.HTTPError as e:
        if e.code == 404:
            print "No such looking glass path: %s\n(%s)" % (
                " ".join(quoted_args), target_url)
        else:
            print "Error code %d: %s" % (e.getcode(), e.read())
        return
    except urllib2.URLError as e:
        print "No server at http://%s:%d : %s" % (options.server,
                                                  options.port, e)



# encoding: utf-8
"""
tcpflags.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
"""

# http://www.iana.org/assignments/tcp-header-flags
class TCPFlags (int):
	FIN    = 0x1
	SYN    = 0x2
	RST  = 0x4
	PUSH   = 0x8
	ACK    = 0x10
	URGENT = 0x20

	def __str__ (self):
		if self == self.FIN:    return 'fin'
		if self == self.SYN:    return 'syn'
		if self == self.RST:    return 'rst'
		if self == self.PUSH:   return 'push'
		if self == self.ACK:    return 'ack'
		if self == self.URGENT: return 'urgent'
		return 'invalid tcp flag %d' % int.__str__(self)

	def __repr__ (self):
		return str(self)

def NamedTCPFlags (name):
	flag = name.lower()
	if flag == 'fin':    return TCPFlags(TCPFlags.FIN)
	if flag == 'syn':    return TCPFlags(TCPFlags.SYN)
	if flag == 'rst':    return TCPFlags(TCPFlags.RST)
	if flag == 'push':   return TCPFlags(TCPFlags.PUSH)
	if flag == 'ack':    return TCPFlags(TCPFlags.ACK)
	if flag == 'urgent': return TCPFlags(TCPFlags.URGENT)
	raise ValueError('invalid flag name %s' % flag)



# encoding: utf-8
"""
attribute/__init__.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

from struct import pack

from bagpipe.exabgp.message.update.attribute.flag import Flag
from bagpipe.exabgp.message.update.attribute.id import AttributeID

# =================================================================== Attribute

class Attribute (object):
	ID   = 0x00
	FLAG = 0x00

	def _attribute (self,value):
		flag = self.FLAG
		if flag & Flag.OPTIONAL and not value:
			return ''
		length = len(value)
		if length > 0xFF:
			flag |= Flag.EXTENDED_LENGTH
		if flag & Flag.EXTENDED_LENGTH:
			len_value = pack('!H',length)
		else:
			len_value = chr(length)
		return "%s%s%s%s" % (chr(flag),chr(self.ID),len_value,value)

	def __eq__ (self,other):
		return self.ID == other.ID




# encoding: utf-8
"""
process.py

Created by Thomas Mangin on 2011-05-02.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import os
import subprocess
import select

from bagpipe.exabgp.log import Logger
logger = Logger()

class ProcessError (Exception):
	pass

class Processes (object):
	def __init__ (self,supervisor):
		self.supervisor = supervisor
		self._process = {}
		self._receive_routes = {}
		self._notify = {}
		self._broken = []

	def _terminate (self,name):
		logger.processes("Terminating process %s" % name)
		self._process[name].terminate()
		self._process[name].wait()
		del self._process[name]

	def terminate (self):
		for name in list(self._process):
			try:
				self._terminate(name)
			except OSError:
				# we most likely received a SIGTERM signal and our child is already dead
				pass

	def _start (self,name):
		try:
			if name in self._process:
				logger.processes("process already running")
				return
			proc = self.supervisor.configuration.process
			if not name in proc:
				logger.processes("Can not start process, no configuration for it (anymore ?)")
				return
			self._receive_routes[name] = proc[name]['receive-routes']
			self._process[name] = subprocess.Popen(proc[name]['run'],
				stdin=subprocess.PIPE,
				stdout=subprocess.PIPE,
				preexec_fn=os.setsid
			)
			neighbor = proc[name]['neighbor']
			self._notify.setdefault(neighbor,[]).append(name)
			logger.processes("Forked process %s" % name)
		except (subprocess.CalledProcessError,OSError,ValueError):
			self._broken.append(name)
			logger.processes("Could not start process %s" % name)

	def start (self):
		proc = self.supervisor.configuration.process
		for name in proc:
			self._start(name)
		for name in list(self._process):
			if not name in proc:
				self._terminate(name)

	def notify (self,neighbor):
		for name in self._notify.get(neighbor,[]):
			yield name
		for name in self._notify.get('*',[]):
			yield name

	def broken (self,neighbor):
		if self._broken:
			for name in self._notify.get(neighbor,[]):
				if name in self._broken:
					return True
			if '*' in self._broken:
				return True
		return False

	def received (self):
		lines = {}
		for name in list(self._process):
			try:
				proc = self._process[name]
				r = True
				while r:
					r,_,_ = select.select([proc.stdout,],[],[],0)
					if r:
						# XXX: readline is blocking, so we are taking the assuption that it will not block
						# XXX: most likely not going to but perhaps the code should be more robust ?
						line = proc.stdout.readline().rstrip()
						if not line:
							# It seems that when we send ^C this is passed to the children to
							# And if they do not intercept it correctly, select.select returns but
							# there is not data to read
							r = False
						else:
							logger.processes("Command from process %s : %s " % (name,line))
							lines.setdefault(name,[]).append(line)
			except (subprocess.CalledProcessError,OSError,ValueError):
				logger.processes("Issue with the process, terminating it and restarting it")
				self._terminate(name)
				self._start(name)
		return lines

	def write (self,name,string):
		while True:
			try:
				self._process[name].stdin.write('%s\r\n' % string)
			except IOError,e:
				self._broken.append(name)
				if e.errno == 32:
					self._broken.append(name)
					logger.processes("Issue while sending data to our helper program")
					raise ProcessError()
				else:
					# Could it have been caused by a signal ? What to do. 
					logger.processes("REPORT TO DEVELOPERS: IOError received while SENDING data to helper program %s, retrying" % str(e.errno))
					continue
			break

		try:
			self._process[name].stdin.flush()
		except IOError,e:
			# AFAIK, the buffer should be flushed at the next attempt.
			logger.processes("REPORT TO DEVELOPERS: IOError received while FLUSHING data to helper program %s, retrying" % str(e.errno))

		return True

	# return all the process which are interrested in route update notification
	def receive_routes (self):
		for name in self._process:
			if self._receive_routes[name]:
				yield name



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import logging
import socket

from bagpipe.bgp.engine import RouteEvent, RouteEntry

from bagpipe.exabgp.structure.address import AFI, SAFI
from bagpipe.exabgp.structure.ip import Inet
from bagpipe.exabgp.message.update.attribute.communities import RouteTarget
from bagpipe.exabgp.message.update.attributes import Attributes
from bagpipe.exabgp.message.update.attribute.nexthop import NextHop
from bagpipe.exabgp.message.update.attribute.localpref import LocalPreference

WAIT_TIME = 0.05

RT1 = RouteTarget(64512, None, 10)
RT2 = RouteTarget(64512, None, 20)
RT3 = RouteTarget(64512, None, 30)

NLRI1 = "NLRI1"
NLRI2 = "NLRI2"

NH1 = Inet(1, socket.inet_pton(socket.AF_INET, "1.1.1.1"))
NH2 = Inet(1, socket.inet_pton(socket.AF_INET, "2.2.2.2"))
NH3 = Inet(1, socket.inet_pton(socket.AF_INET, "3.3.3.3"))

NBR = "NBR"
BRR = "BRR"

logging.basicConfig(level=logging.DEBUG,
                    filename="bagpipe-bgp-testsuite.log",
                    format="%(asctime)s %(threadName)-30s %(name)-30s "
                    "%(levelname)-8s %(message)s")

log = logging.getLogger()


class BaseTestBagPipeBGP():

    def setEventTargetWorker(self, worker):
        self.eventTargetWorker = worker

    def _newRouteEvent(self, eventType, nlri, rts, source, nh, lp=0,
                       replacedRouteEntry=None,
                       afi=AFI(AFI.ipv4), safi=SAFI(SAFI.mpls_vpn)):
        attributes = Attributes()
        attributes.add(NextHop(nh))
        attributes.add(LocalPreference(lp))
        routeEvent = RouteEvent(eventType, RouteEntry(
            afi, safi, rts, nlri, attributes, source), source)
        routeEvent.setReplacedRoute(replacedRouteEntry)

        self.eventTargetWorker.enqueue(routeEvent)

        log.info("Emitting event to %s: %s",
                 self.eventTargetWorker, routeEvent)

        self._wait()

        return routeEvent

    def _wait(self):
        time.sleep(WAIT_TIME)

    def _append_call(self, obj):
        log.info("****** %s ******", obj)
        self._calls.append(obj)



#!/usr/bin/env python
#
# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

import re
import urllib2
import json

from optparse import OptionParser
from copy import copy

from netaddr.ip import IPNetwork

from bagpipe.bgp.common.run_command import runCommand
from bagpipe.bgp.common.net_utils import get_device_mac

import logging

DEFAULT_VPN_INSTANCE_ID = "bagpipe-test"

NS2VPN_INTERFACE_PREFIX = "tovpn-"
VPN2NS_INTERFACE_PREFIX = "tons-"

NS2VPN_DEFAULT_IFNAME = "tovpn"


LINUX_DEV_LEN = 15

# Needed so that the OVS bridge kernel interface can hava a high enough MTU
DEFAULT_MTU = 9000

logFormatter = logging.Formatter("[%(levelname)-5.5s]  %(message)s")
log = logging.getLogger()

consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)
log.addHandler(consoleHandler)

log.setLevel(logging.WARNING)


def create_veth_pair(vpn_interface, ns_interface, ns_name):
    runCommand(log, "ip netns exec %s ip link delete %s" % (ns_name,
                                                            ns_interface),
               raiseExceptionOnError=False,
               acceptableReturnCodes=[0, 1])
    # in case the interface was previously attached to OVS,
    # we need to remove that, or when then interface is re-created
    # OVS will take it back!
    runCommand(log, "ovs-vsctl del-port %s" % vpn_interface,
               raiseExceptionOnError=False,
               acceptableReturnCodes=[0, 1])
    runCommand(log, "ip link delete %s" % vpn_interface,
               raiseExceptionOnError=False,
               acceptableReturnCodes=[0, 1])
    runCommand(log, "ip link add %s type veth peer name %s netns %s" %
               (vpn_interface, ns_interface, ns_name),
               raiseExceptionOnError=False)
    runCommand(log, "ip link set dev %s up" % vpn_interface)
    runCommand(log, "ip link set dev %s mtu %d" % (vpn_interface, DEFAULT_MTU))
    runCommand(log, "ip netns exec %s ip link set dev %s up" %
               (ns_name, ns_interface))


def get_vpn2ns_if_name(namespace):
    return (VPN2NS_INTERFACE_PREFIX + namespace)[:LINUX_DEV_LEN]


def createSpecialNetNSPort(options):
    print "Will plug local namespace %s into network" % options.netns

    # create namespace
    runCommand(log, "ip netns add %s" %
               options.netns, raiseExceptionOnError=False)

    # create veth pair and move one into namespace
    if options.ovs_vlan:
        create_veth_pair(options.if2netns, "ns2vpn-raw", options.netns)

        runCommand(log, "ip netns exec %s ip link add link ns2vpn-raw "
                   "name %s type vlan id %d"
                   % (options.netns, options.if2vpn, options.ovs_vlan))
        runCommand(log, "ip netns exec %s ip link set %s up"
                   % (options.netns, options.if2vpn))
    else:
        create_veth_pair(options.if2netns, options.if2vpn, options.netns)

    if options.mac:
        runCommand(log, "ip netns exec %s ip link set %s address %s"
                   % (options.netns, options.if2vpn, options.mac))

    runCommand(log, "ip netns exec %s ip addr add %s dev %s" %
               (options.netns, options.ip, options.if2vpn),
               raiseExceptionOnError=False)

    runCommand(log, "ip netns exec %s ip route add default dev %s via %s" %
               (options.netns, options.if2vpn, options.gw_ip),
               raiseExceptionOnError=False)

    runCommand(log, "ip netns exec %s ip link set %s mtu 1420" %
               (options.netns, options.if2vpn),
               raiseExceptionOnError=False)


def main():
    usage = "usage: %prog [--attach|--detach] --network-type (ipvpn|evpn) "\
        "--port (<port>|netns) --ip <ip>[/<mask>] [options] (see --help)"
    parser = OptionParser(usage)

    parser.add_option("--attach", dest="operation",
                      action="store_const", const="attach",
                      help="attach local port")
    parser.add_option("--detach", dest="operation",
                      action="store_const", const="detach",
                      help="detach local port")

    parser.add_option("--network-type", dest="network_type",
                      help="network type (ipvpn or evpn)",
                      choices=["ipvpn", "evpn"])
    parser.add_option("--vpn-instance-id", dest="vpn_instance_id",
                      help="UUID for the network instance "
                      "(default: %default-(ipvpn|evpn))",
                      default=DEFAULT_VPN_INSTANCE_ID)
    parser.add_option("--port", dest="port",
                      help="local port to attach/detach (use special port "
                      "'netns[:if]' to have an interface to a local network "
                      "namespace attached/detached "
                      "[with 'if' as the name of the interface to the netns]")

    parser.add_option("--rt", dest="routeTargets",
                      help="route target [default: 64512:0] (can be "
                      "specified multiple times)", default=[], action="append")
    parser.add_option("--import-rt", dest="importOnlyRouteTargets",
                      help="import-only route target (can be specified"
                      "multiple times)", default=[], action="append")
    parser.add_option("--export-rt", dest="exportOnlyRouteTargets",
                      help="export-only route target (can be specified"
                      "multiple times)", default=[], action="append")

    parser.add_option("--ip", dest="ip",
                      help="IP prefix / mask (mask defaults to /24)")
    parser.add_option("--gateway-ip", dest="gw_ip",
                      help="IP address of network gateway (optional, "
                      "defaults to last IP in range)")
    parser.add_option("--mac", dest="mac",
                      help="MAC address (required for evpn if port"
                      " is not 'netns')")

    parser.set_defaults(advertiseSubnet=False)
    parser.add_option("--advertise-singleton", action="store_false",
                      dest="advertiseSubnet",
                      help="advertise IP address as a /32 (default)")

    parser.add_option("--advertise-subnet", action="store_true",
                      dest="advertiseSubnet",
                      help="advertise the whole IP subnet")

    parser.add_option("--ovs-preplug", action="store_true", dest="ovs_preplug",
                      default=False, help="should we prealably plug the port "
                      "into an OVS bridge")
    parser.add_option("--ovs-bridge", dest="bridge", default="br-int",
                      help="if preplug, specifies which OVS bridge to use"
                      " (default: %default)")
    parser.add_option("--ovs-vlan", dest="ovs_vlan", type='int',
                      help="if specified, only this VLAN from the OVS "
                      "interface will be attached to the VPN instance "
                      "(optional)")

    parser.add_option("--netns", dest="netns",
                      help="name of network namespace (optional, for use with"
                      " --port netns)")
    parser.add_option("--if2vpn", dest="if2vpn", default=NS2VPN_DEFAULT_IFNAME,
                      help="name of interface in netns toward VPN"
                      "defaults to %default "
                      "(optional, for use with --port netns)")

    parser.add_option("--readv-from-rt", dest="reAdvFromRTs",
                      help="enables route readvertisement from these RTs,"
                      " works in conjunction with --readv-to-rt",
                      default=[], action="append")

    parser.add_option("--readv-to-rt", dest="reAdvToRTs",
                      help="enables route readvertisement to these RTs,"
                      " works in conjunction with --readv-from-rt",
                      default=[], action="append")

    (options, _) = parser.parse_args()

    if not(options.operation):
        parser.error("Need to specify --attach or --detach")

    if not(options.port):
        parser.error("Need to specify --port <localport>")

    if not(options.network_type):
        parser.error("Need to specify --network-type")

    if not(options.ip):
        parser.error("Need to specify --ip")

    if (len(options.routeTargets) == 0 and
            not (options.importOnlyRouteTargets
                 or options.exportOnlyRouteTargets)):
        if options.network_type == "ipvpn":
            options.routeTargets = ["64512:512"]
        else:
            options.routeTargets = ["64512:513"]

    importRTs = copy(options.routeTargets or [])
    for rt in options.importOnlyRouteTargets:
        importRTs.append(rt)

    exportRTs = copy(options.routeTargets or [])
    for rt in options.exportOnlyRouteTargets:
        exportRTs.append(rt)

    if not re.match('.*/[0-9]+$', options.ip):
        options.ip = options.ip + "/24"

    if not(options.gw_ip):
        net = IPNetwork(options.ip)
        print "using %s as gateway address" % str(net[-2])
        options.gw_ip = str(net[-2])

    if options.vpn_instance_id == DEFAULT_VPN_INSTANCE_ID:
        options.vpn_instance_id = "%s-%s" % (
            options.network_type, options.vpn_instance_id)

    if options.port.startswith("netns"):

        if not options.netns:
            options.netns = options.vpn_instance_id

        try:
            (_, options.if2netns) = options.port.split(":")
        except:
            options.if2netns = get_vpn2ns_if_name(options.netns)

        if options.operation == "attach":
            createSpecialNetNSPort(options)

        options.port = options.if2netns
        if not options.mac:
            options.mac = get_device_mac(lambda *args: runCommand(log, *args),
                                         options.if2vpn, options.netns)

        print "Local port: %s (%s)" % (options.port, options.mac)
        runCommand(log, "ip link show %s" % options.port)

    local_port = {}
    if options.port[:5] == "evpn:":
        if (options.network_type == "ipvpn"):
            print "will plug evpn %s into the IPVPN" % options.port[5:]
            local_port['evpn'] = {'id': options.port[5:]}
        else:
            raise Exception("Can only plug an evpn into an ipvpn")
    else:
        local_port['linuxif'] = options.port

        # currently our only the MPLS OVS driver for ipvpn requires preplug
        if (options.ovs_preplug and options.network_type == "ipvpn"):
            print "pre-plugging %s into %s" % (options.port,
                                               options.bridge)
            runCommand(log, "ovs-vsctl del-port %s %s" %
                       (options.bridge, options.port),
                       raiseExceptionOnError=False)
            runCommand(log, "ovs-vsctl add-port %s %s" %
                       (options.bridge, options.port))

            local_port['ovs'] = {'port_name': options.port,
                                 'plugged': True}

            if options.ovs_vlan:
                local_port['ovs']['vlan'] = options.ovs_vlan

    if not options.mac:
        if options.network_type == IPVPN:
            options.mac = "52:54:00:99:99:22"
        else:
            parser.error("Need to specify --mac for an EVPN network "
                         "attachment if port is not 'netns'")

    readvertise = None
    if options.reAdvToRTs:
        readvertise = {"from_rt": options.reAdvFromRTs,
                       "to_rt": options.reAdvToRTs}

    json_data = json.dumps({"import_rt":  importRTs,
                            "export_rt":  exportRTs,
                            "local_port":  local_port,
                            "vpn_instance_id":  options.vpn_instance_id,
                            "vpn_type":    options.network_type,
                            "gateway_ip":  options.gw_ip,
                            "mac_address": options.mac,
                            "ip_address":  options.ip,
                            "advertise_subnet": options.advertiseSubnet,
                            "readvertise": readvertise
                            })

    print "request: %s" % json_data

    os.environ['NO_PROXY'] = "127.0.0.1"
    req = urllib2.Request("http://127.0.0.1:8082/%s_localport" %
                          options.operation, json_data,
                          {'Content-Type': 'application/json'})
    try:
        response = urllib2.urlopen(req)
        response_content = response.read()
        response.close()

        print "response: %d %s" % (response.getcode(), response_content)
    except urllib2.HTTPError as e:
        error_content = e.read()
        print "   %s" % error_content
        sys.exit("error %d, reason: %s" % (e.code, e.reason))



# encoding: utf-8
"""
ip.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2010-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

import math
import socket

from bagpipe.exabgp.structure.address import AFI,SAFI
from bagpipe.exabgp.message.update.route import Route

_bgp = {}
for netmask in range(0,129):
	_bgp[netmask] = int(math.ceil(float(netmask)/8))

def _detect_afi(ip):
	if ip.count(':'):
		return AFI.ipv6
	return AFI.ipv4

def to_IP (ip):
	afi = _detect_afi(ip)
	af = Inet._af[afi]
	network = socket.inet_pton(af,ip)
	return Inet(afi,network)

def to_Route (ip,mask):
	afi = _detect_afi(ip)
	return Route(Prefix(afi,ip,mask))

class Inet (object):
	_UNICAST = SAFI(SAFI.unicast)
	_MULTICAST = SAFI(SAFI.multicast)
	_unicast_range = set(range(224,240)) # 239 is last

	"""An IP in the 4 bytes format"""
	# README: yep, we should surely change this _ name here
	_af = {
		AFI.ipv4: socket.AF_INET,
		AFI.ipv6: socket.AF_INET6,
	}

	_afi = {
		socket.AF_INET : AFI.ipv4,
		socket.AF_INET6: AFI.ipv6,
	}

	_length = {
		AFI.ipv4:  4,
		AFI.ipv6: 16,
	}

	def __init__ (self,afi,raw):
		self.afi = AFI(afi)
		self.raw = raw
		self.__update()

	def __update (self):
		self.ip = self._ip()

		if self.afi == AFI.ipv4 and int(self.ip.split('.')[0]) in self._unicast_range:
			self.safi = self._MULTICAST
		else:
			self.safi = self._UNICAST

	def update_raw (self,raw):
		self.raw = raw
		self.__update()

	def pack (self):
		return self.raw

	def _ip (self):
		try:
			return socket.inet_ntop(self._af[self.afi],self.raw)
		except socket.error:
			raise ValueError('invalid IP')

	def __len__ (self):
		return len(self.raw)

	def __str__ (self):
		return self.ip

	def __repr__ (self):
		return str(self)

	def __eq__ (self,other):
		return self.raw == other.raw and self.safi == other.safi

class _Prefix (Inet):
	# have a .raw for the ip
	# have a .mask for the mask
	# have a .bgp with the bgp wire format of the prefix

	def __init__(self,af,ip,mask):
		self.mask = int(mask)
		Inet.__init__(self,af,ip)

	def __str__ (self):
		return "%s/%s" % (self.ip,self.mask)

	def __repr__ (self):
		return str(self)

	def pack (self):
		return chr(self.mask) + self.raw[:_bgp[self.mask]]

	def __eq__(self,other):
		return isinstance(other,_Prefix) and self.pack() == other.pack()

	def __hash__(self):
		return hash(self.pack())

	def __len__ (self):
		return _bgp[self.mask] + 1


class BGPPrefix (_Prefix):
	"""From the BGP prefix wire format, Store an IP (in the network format), its netmask and the bgp format"""
	def __init__ (self,afi,bgp):
		end = _bgp[ord(bgp[0])]
		_Prefix.__init__(self,afi,bgp[1:end+1] + '\0'*(self._length[afi]-end),ord(bgp[0]))

class AFIPrefix (_Prefix):
	"""Store an IP (in the network format), its netmask and the bgp format of the IP"""
	def __init__ (self,afi,network,mask):
		_Prefix.__init__(self,afi,network,mask)

class Prefix (_Prefix):
	"""Store an IP (in the network format), its netmask and the bgp format of the IP"""
	def __init__ (self,afi,ip,mask):
		af = self._af[afi]
		network = socket.inet_pton(af,ip)
		_Prefix.__init__(self,afi,network,mask)



# encoding: utf-8
"""
utils.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
"""

import os
import sys
import time
import logging
import logging.handlers

class LazyFormat (object):
	def __init__ (self,prefix,format,message):
		self.prefix = prefix
		self.format = format
		self.message = message

	def __str__ (self):
		if self.format:
			return self.prefix + self.format(self.message)
		return self.prefix + self.message

	def split (self,c):
		return str(self).split(c)

class _Logger (object):
	_instance = None
	_syslog = None

	_history = []
	_max_history = 20

	_config = ''
	_pid = os.getpid()

	# we use os.pid everytime as we may fork and the class is instance before it

	def config (self,config=None):
		if config is not None:
			self._config = config
		return self._config

	def history (self):
		return "\n".join(self._format(*_) for _ in self._history)

	def _record (self,timestamp,level,source,message):
		if len(self._history) > self._max_history:
			self._history.pop(0)
		self._history.append((timestamp,level,source,message))

	def _format (self,timestamp,level,source,message):
		now = time.strftime('%a, %d %b %Y %H:%M:%S',timestamp)
		return "%s %-8s %-6d %-13s %s" % (now,level,self._pid,source,message)

	def _prefixed (self,level,source,message):
		ts = time.localtime()
		self._record(ts,level,source,message)
		return self._format(ts,level,source,message)

	def __init__ (self):
		if os.environ.get('DEBUG_SUPERVISOR','1') in ['1','yes','Yes','YES']: self._supervisor = True
		else: self._supervisor = False

		if os.environ.get('DEBUG_DAEMON','1') in ['1','yes','Yes','YES']: self._daemon = True
		else: self._daemon = False

		if os.environ.get('DEBUG_PROCESSES','1') in ['1','yes','Yes','YES']: self._processes = True
		else: self._processes = False

		if os.environ.get('DEBUG_CONFIGURATION','0') == '0': self._configuration = False
		else: self._configuration = True

		if os.environ.get('DEBUG_WIRE','0') == '0': self._wire = False
		else: self._wire = True

		if os.environ.get('DEBUG_MESSAGE','0') in ['1','yes','Yes','YES']: self._message = True
		else: self._message = False

		if os.environ.get('DEBUG_RIB','0') == '0': self._rib = False
		else: self._rib = True

		if os.environ.get('DEBUG_TIMER','0') == '0': self._timer = False
		else: self._timer = True

		if os.environ.get('DEBUG_ROUTE','0') == '0': self._routes = False
		else: self._routes = True

		# DEPRECATED, kept for compatibility in 2.0.x series
		if os.environ.get('DEBUG_ROUTES','0') == '0': self._routes = False
		else: self._routes = True

		if os.environ.get('DEBUG_PARSER','0') == '0': self._parser = False
		else: self._parser = True

		if not os.environ.get('DEBUG_ALL','0') == '0':
			self._supervisor = True
			self._daemon = True
			self._processes = True
			self._configuration = True
			self._wire = True
			self._message = True
			self._rib = True
			self._timer = True
			self._routes = True
			self._parser = True

		if not os.environ.get('DEBUG_CORE','0') == '0':
			self._supervisor = True
			self._daemon = True
			self._processes = True
			#self._configuration = True
			#self._wire = True
			self._message = True
			#self._rib = True
			self._timer = True
			self._routes = True
			self._parser = False

		destination = os.environ.get('SYSLOG',None)
		if destination is None:
			return

		try:
			if destination == '':
				if sys.platform == "darwin":
					address = '/var/run/syslog'
				else:
					address = '/dev/log'
				if not os.path.exists(address):
					address = ('localhost', 514)
				handler = logging.handlers.SysLogHandler(address)
			elif destination.lower().startswith('host:'):
				# If the address is invalid, each syslog call will print an error.
				# See how it can be avoided, as the socket error is encapsulated and not returned
				address = (destination[5:].strip(), 514)
				handler = logging.handlers.SysLogHandler(address)
			else:
				handler = logging.handlers.RotatingFileHandler(destination, maxBytes=5*1024*1024, backupCount=5)
			self._syslog = logging.getLogger()
			self._syslog.setLevel(logging.DEBUG)
			self._syslog.addHandler(handler)
		except IOError:
			self.critical('Can not use SYSLOG, failing back to stdout')

	def debug (self,message,source='',level='DEBUG'):
		for line in message.split('\n'):
			if self._syslog:
				self._syslog.debug(self._prefixed(level,source,line))
			else:
				print self._prefixed(level,source,line)
				sys.stdout.flush()

	def info (self,message,source='',level='INFO'):
		for line in message.split('\n'):
			if self._syslog:
				self._syslog.info(self._prefixed(level,source,line))
			else:
				print self._prefixed(level,source,line)
				sys.stdout.flush()

	def warning (self,message,source='',level='WARNING'):
		for line in message.split('\n'):
			if self._syslog:
				self._syslog.warning(self._prefixed(level,source,line))
			else:
				print self._prefixed(level,source,line)
				sys.stdout.flush()

	def error (self,message,source='',level='ERROR'):
		for line in message.split('\n'):
			if self._syslog:
				self._syslog.error(self._prefixed(level,source,line))
			else:
				print self._prefixed(level,source,line)
				sys.stdout.flush()

	def critical (self,message,source='',level='CRITICAL'):
		for line in message.split('\n'):
			if self._syslog:
				self._syslog.critical(self._prefixed(level,source,line))
			else:
				print self._prefixed(level,source,line)
				sys.stdout.flush()

	# show the message on the wire
	def wire (self,message):
		if self._wire:
			self.debug(message,'wire')
		else:
			self._record(time.localtime(),'wire','DEBUG',message)

	# show the exchange of message between peers
	def message (self,message):
		if self._message:
			self.info(message,'message')
		else:
			self._record(time.localtime(),'message','info',message)

	# show the parsing of the configuration
	def configuration (self,message):
		if self._configuration:
			self.info(message,'configuration')
		else:
			self._record(time.localtime(),'configuration','info',message)

	# show the exchange of message generated by the supervisor (^C and signal received)
	def supervisor (self,message):
		if self._supervisor:
			self.info(message,'supervisor')
		else:
			self._record(time.localtime(),'supervisor','info',message)

	# show the change of rib table
	def rib (self,message):
		if self._rib:
			self.info(message,'rib')
		else:
			self._record(time.localtime(),'rib','info',message)

	# show the change of rib table
	def timers (self,message):
		if self._timer:
			self.info(message,'timers')
		else:
			self._record(time.localtime(),'timers','info',message)

	# show the exchange of message generated by the daemon feature (change pid, fork, ...)
	def daemon (self,message):
		if self._daemon:
			self.info(message,'daemon')
		else:
			self._record(time.localtime(),'daemon','info',message)

	# show the exchange of message generated by the forked processes
	def processes (self,message):
		if self._processes:
			self.info(message,'processes')
		else:
			self._record(time.localtime(),'processes','info',message)

	# show the exchange of message generated by the routes received
	def routes (self,message):
		if self._routes:
			self.info(message,'route')
		else:
			self._record(time.localtime(),'route','info',message)

	# show how the message received are parsed
	def parser (self,message):
		if self._parser:
			self.info(message,'parser')

def Logger ():
	if _Logger._instance:
		return _Logger._instance
	instance = _Logger()
	_Logger._instance = instance
	return instance

if __name__ == '__main__':
	logger = Logger()
	logger.wire('wire packet content')
	logger.message('message exchanged')
	logger.debug('debug test')




# encoding: utf-8
"""
Copyright (c) 2014, Orange
All rights reserved.

File released under the BSD 3-Clause license.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions 
are met:

1. Redistributions of source code must retain the above copyright 
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in 
   the documentation and/or other materials provided with the 
   distribution.

3. Neither the name of the copyright holder nor the names of its 
   contributors may be used to endorse or promote products derived 
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
"""

from struct import pack

import socket

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

from bagpipe.exabgp.structure.mpls import LabelStackEntry, NO_LABEL



tunnel_types_to_class = dict()
def register(myclass):
    tunnel_types_to_class[myclass.subtype] = myclass


class PMSITunnel(Attribute):
    '''
    http://tools.ietf.org/html/rfc6514#section-5
    
      +---------------------------------+
      |  Flags (1 octet)                |
      +---------------------------------+
      |  Tunnel Type (1 octets)         |
      +---------------------------------+
      |  MPLS Label (3 octets)          |
      +---------------------------------+
      |  Tunnel Identifier (variable)   |
      +---------------------------------+
    '''
    ID = AttributeID.PMSI_TUNNEL
    FLAG = Flag.OPTIONAL | Flag.TRANSITIVE
    MULTIPLE = False

    def __init__ (self,subtype,label=NO_LABEL,flags=0,packedTunnelId=None):
        if label==None: label=NO_LABEL
        if not isinstance(label,LabelStackEntry): raise Exception("label should be of LabelStackEntry type (is: %s)" % type(label))
        self.subtype = subtype
        self.label = label
        self.pmsi_flags = flags
        self.packedTunnelId = packedTunnelId

    def pack(self):
        if self.packedTunnelId is None:
            self._computePackedTunnelId()
        return self._attribute(pack('!BB', self.pmsi_flags, self.subtype) + self.label.pack() + self.packedTunnelId)
    
    def __len__ (self):
        if self.packedTunnelId is None:
            self._computePackedTunnelId()
        return 4+len(self.packedTunnelId)

    def __str__ (self):
        if self.subtype in tunnel_types_to_class:
            type_string = tunnel_types_to_class[self.subtype].nickname
            return "PMSITunnel:%s:%s:[%s]" % (type_string,str(self.pmsi_flags) or '',self.label or "-")
        else:
            type_string = "%d" % self.subtype
            return "PMSITunnel:%s:%s:[%s]:%s" % (type_string,str(self.pmsi_flags) or '',self.label or "-","xxx") 
            #TODO: add hex dump of packedValue 

    def __repr__ (self):
        return str(self)

    def _computePackedTunnelId(self):
        raise Exception("Abstract class, cannot compute packedTunnelId")
    
    def __cmp__(self,o):
        if (not isinstance(o,PMSITunnel) or
            (self.subtype != o.subtype) or
            (self.label != o.label) or
            (self.pmsi_flags != o.pmsi_flags) or
            (self.packedTunnelId != o.packedTunnelId)
            ): 
            return -1
        else:
            return 0
    
    @staticmethod
    def unpack(data):
        #flags
        flags = ord(data[0])
        data=data[1:]
        
        #subtype
        subtype = ord(data[0])
        data=data[1:]
        
        #label
        label = LabelStackEntry.unpack(data[:3])
        data=data[3:]
        
        if subtype in tunnel_types_to_class:
            return tunnel_types_to_class[subtype].unpack(label,flags,data)
        else:
            return PMSITunnel(subtype,label,flags,data)



class PMSITunnelIngressReplication(PMSITunnel):

    subtype = 6
    nickname = "IngressReplication"

    def __init__(self,ip,label=NO_LABEL,flags=0):
        try:
            socket.inet_pton( socket.AF_INET, ip )
        except:
            raise Exception("Malformed IP address")
        self.ip = ip

        PMSITunnel.__init__(self, self.subtype, label, flags)
        
    def __str__ (self):
        desc = "[%s]" % self.ip
        return PMSITunnel.__str__(self) + ":" + desc
        
    def _computePackedTunnelId(self):
        self.packedTunnelId = socket.inet_pton( socket.AF_INET, self.ip )

    @staticmethod
    def unpack(label,flags,data):
        ip = socket.inet_ntop( socket.AF_INET, data[:4] )
        return PMSITunnelIngressReplication(ip,label,flags)


register(PMSITunnelIngressReplication)




# vim: tabstop=4 shiftwidth=4 softtabstop=4
# encoding: utf-8

# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

import socket

from bagpipe.bgp.common import logDecorator

from bagpipe.bgp.common.looking_glass import LookingGlassLocalLogger, LGMap

from bagpipe.bgp.common.run_command import runCommand

from bagpipe.exabgp.message.update.attribute.communities import Encapsulation


class DataplaneDriver(LookingGlassLocalLogger):
    __metaclass__ = ABCMeta

    dataplaneInstanceClass = None

    encaps = [Encapsulation(Encapsulation.DEFAULT)]
    makeB4BreakSupport = False
    ecmpSupport = False

    @logDecorator.log
    def __init__(self, config, init=True):
        '''config is a dict'''
        LookingGlassLocalLogger.__init__(self)

        assert(issubclass(self.dataplaneInstanceClass, VPNInstanceDataplane))

        self.config = config

        self.local_address = None
        try:
            self.local_address = self.config["dataplane_local_address"]
            socket.inet_pton(socket.AF_INET, self.local_address)
            self.log.info("Will use %s as local_address", self.local_address)
        except KeyError:
            self.log.info("Will use BGP address as dataplane_local_address")
            self.local_address = None
        except socket.error:
            raise Exception("malformed local_address: '%s'" %
                            self.local_address)

        # skipped if instantiated with init=False, to be used for cleanup
        if init:
            self._initReal(config)

        # Flag to trigger cleanup all dataplane states on first call to
        # vifPlugged
        self.firstInit = True

    @abstractmethod
    def resetState(self):
        pass

    @abstractmethod
    def _initReal(self, config):
        '''
        This is called after resetState (which, e.g. cleans up the stuff
        possibly left-out by a previous failed run).

        All init things that should not be cleaned up go here.
        '''
        pass

    @logDecorator.logInfo
    def initializeDataplaneInstance(self, instanceId, externalInstanceId,
                                    gatewayIP, mask, instanceLabel, **kwargs):
        '''
        returns a VPNInstanceDataplane subclass
        after calling resetState on the dataplane driver, if this is the first
        call to initializeDataplaneInstance
        '''

        if self.firstInit:
            self.log.info("First VPN instance init, resetting dataplane state")
            try:
                self.resetState()
            except Exception as e:
                self.log.error("Exception while resetting state: %s", e)
            self.firstInit = False
        else:
            self.log.debug("(not resetting dataplane state)")

        return self.dataplaneInstanceClass(self, instanceId,
                                           externalInstanceId, gatewayIP, mask,
                                           instanceLabel, **kwargs)

    def cleanup(self):
        # FIXME: to be clarified: can be removed ? should call resetState ?
        self._cleanupReal()

    def getLocalAddress(self):
        return self.local_address

    def supportedEncaps(self):
        return self.__class__.encaps

    def _runCommand(self, command, *args, **kwargs):
        return runCommand(self.log, command, *args, **kwargs)

    def getLGMap(self):
        encaps = []
        for encap in self.supportedEncaps():
            encaps.append(repr(encap))
        return {
            "name": (LGMap.VALUE, self.__class__.__name__),
            "local_address": (LGMap.VALUE, self.local_address),
            "supported_encaps": (LGMap.VALUE, encaps),
            "config": (LGMap.VALUE, self.config)
        }


class VPNInstanceDataplane(LookingGlassLocalLogger):
    __metaclass__ = ABCMeta

    @logDecorator.logInfo
    def __init__(self, dataplaneDriver, instanceId, externalInstanceId,
                 gatewayIP, mask, instanceLabel=None):
        LookingGlassLocalLogger.__init__(self, repr(instanceId))
        self.driver = dataplaneDriver
        self.config = dataplaneDriver.config
        self.instanceId = instanceId
        self.externalInstanceId = externalInstanceId
        self.gatewayIP = gatewayIP
        self.mask = mask
        self.instanceLabel = instanceLabel

    @abstractmethod
    def cleanup(self):
        pass

    @abstractmethod
    def vifPlugged(self, macAddress, ipAddressPrefix, localPort, label):
        pass

    @abstractmethod
    def vifUnplugged(self, macAddress, ipAddressPrefix, localPort, label,
                     lastEndpoint=True):
        pass

    def update_fallback(self, fallback):
        if fallback is not None:
            self.log.warning("fallback  specified (%s) but not supported by"
                             " driver, ignoring", fallback)

    @abstractmethod
    def setupDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri,
                                        encaps):
        pass

    @abstractmethod
    def removeDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri):
        pass

    def _runCommand(self, *args, **kwargs):
        return runCommand(self.log, *args, **kwargs)

    # Looking glass info ####

    def getLGMap(self):
        return {
            "driver": (LGMap.DELEGATE, self.driver),
        }


class DummyVPNInstanceDataplane(VPNInstanceDataplane):

    @logDecorator.log
    def __init__(self, *args, **kwargs):
        VPNInstanceDataplane.__init__(self, *args)

    @logDecorator.log
    def vifPlugged(self, macAddress, ipAddressPrefix, localPort, label):
        pass

    @logDecorator.log
    def vifUnplugged(self, macAddress, ipAddressPrefix, localPort, label,
                     lastEndpoint=True):
        pass

    @logDecorator.log
    def update_fallback(self, fallback):
        pass

    @logDecorator.log
    def setupDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri,
                                        encaps):
        pass

    @logDecorator.log
    def removeDataplaneForRemoteEndpoint(self, prefix, remotePE, label, nlri):
        pass

    @logDecorator.log
    def cleanup(self):
        pass


class DummyDataplaneDriver(DataplaneDriver):

    dataplaneInstanceClass = DummyVPNInstanceDataplane

    def __init__(self, *args):
        DataplaneDriver.__init__(self, *args)

    @logDecorator.logInfo
    def _initReal(self, config):
        pass

    @logDecorator.logInfo
    def resetState(self):
        pass

    @logDecorator.logInfo
    def _cleanupReal(self):
        pass



# encoding: utf-8
"""
attributes.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2012 Exa Networks. All rights reserved.
Modified by Orange - 2014
"""

from bagpipe.exabgp.message.update.attribute import AttributeID,Flag,Attribute

# =================================================================== Origin (1)

class Origin (Attribute):
	ID = AttributeID.ORIGIN
	FLAG = Flag.TRANSITIVE
	MULTIPLE = False

	IGP        = 0x00
	EGP        = 0x01
	INCOMPLETE = 0x02

	def __init__ (self,origin):
		self.origin = origin

	def pack (self):
		return self._attribute(chr(self.origin))

	def __len__ (self):
		return len(self.pack())

	def __str__ (self):
		if self.origin == 0x00: return 'IGP'
		if self.origin == 0x01: return 'EGP'
		if self.origin == 0x02: return 'INCOMPLETE'
		return 'INVALID'

	def __repr__ (self):
		return str(self)
	
	def __cmp__(self,other):
		if ( not isinstance(other,Origin) 
			  or (self.origin != other.origin)
		    ):
			return -1
		else:
			return 0
